# The MIT License (MIT)
#
# Copyright (C) 2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .analysis import Analysis, PeriodData


class _PeriodData(PeriodData):
    def __init__(self):
        # Indexed by irq 'id' (irq or vec)
        self.hard_irq_stats = {}
        self.softirq_stats = {}
        # Log of individual interrupts
        self.irq_list = []


class IrqAnalysis(Analysis):
    def __init__(self, state, conf):
        notification_cbs = {
            'irq_handler_entry': self._process_irq_handler_entry,
            'irq_handler_exit': self._process_irq_handler_exit,
            'softirq_exit': self._process_softirq_exit
        }
        super().__init__(state, conf, notification_cbs)

    def _create_period_data(self):
        return _PeriodData()

    def _process_irq_handler_entry(self, period_data, **kwargs):
        id = kwargs['id']
        name = kwargs['irq_name']
        if id not in period_data.hard_irq_stats:
            period_data.hard_irq_stats[id] = HardIrqStats(name)
        elif name not in period_data.hard_irq_stats[id].names:
            period_data.hard_irq_stats[id].names.append(name)

    def _process_irq_handler_exit(self, period_data, **kwargs):
        irq = kwargs['hard_irq']

        if not self._filter_cpu(irq.cpu_id):
            return

        if self._conf.min_duration is not None and \
           irq.duration < self._conf.min_duration:
            return
        if self._conf.max_duration is not None and \
           irq.duration > self._conf.max_duration:
            return

        period_data.irq_list.append(irq)
        if irq.id not in period_data.hard_irq_stats:
            period_data.hard_irq_stats[irq.id] = HardIrqStats()

        period_data.hard_irq_stats[irq.id].update_stats(irq)

    def _process_softirq_exit(self, period_data, **kwargs):
        irq = kwargs['softirq']

        if not self._filter_cpu(irq.cpu_id):
            return

        if self._conf.min_duration is not None and \
           irq.duration < self._conf.min_duration:
            return
        if self._conf.max_duration is not None and \
           irq.duration > self._conf.max_duration:
            return

        period_data.irq_list.append(irq)
        if irq.id not in period_data.softirq_stats:
            name = SoftIrqStats.names[irq.id]
            period_data.softirq_stats[irq.id] = SoftIrqStats(name)

        period_data.softirq_stats[irq.id].update_stats(irq)


class IrqStats():
    def __init__(self, name):
        self._name = name
        self.min_duration = None
        self.max_duration = None
        self.total_duration = 0
        self.irq_list = []

    @property
    def name(self):
        return self._name

    @property
    def count(self):
        return len(self.irq_list)

    def update_stats(self, irq):
        if self.min_duration is None or irq.duration < self.min_duration:
            self.min_duration = irq.duration

        if self.max_duration is None or irq.duration > self.max_duration:
            self.max_duration = irq.duration

        self.total_duration += irq.duration
        self.irq_list.append(irq)

    def reset(self):
        self.min_duration = None
        self.max_duration = None
        self.total_duration = 0
        self.irq_list = []


class HardIrqStats(IrqStats):
    NAMES_SEPARATOR = ', '

    def __init__(self, name='unknown'):
        super().__init__(name)
        self.names = [name]

    @property
    def name(self):
        return self.NAMES_SEPARATOR.join(self.names)


class SoftIrqStats(IrqStats):
    # from include/linux/interrupt.h
    names = {0: 'HI_SOFTIRQ',
             1: 'TIMER_SOFTIRQ',
             2: 'NET_TX_SOFTIRQ',
             3: 'NET_RX_SOFTIRQ',
             4: 'BLOCK_SOFTIRQ',
             5: 'BLOCK_IOPOLL_SOFTIRQ',
             6: 'TASKLET_SOFTIRQ',
             7: 'SCHED_SOFTIRQ',
             8: 'HRTIMER_SOFTIRQ',
             9: 'RCU_SOFTIRQ'}

    def __init__(self, name):
        super().__init__(name)
        self.min_raise_latency = None
        self.max_raise_latency = None
        self.total_raise_latency = 0
        self.raise_count = 0

    def update_stats(self, irq):
        super().update_stats(irq)

        if irq.raise_ts is None:
            return

        raise_latency = irq.begin_ts - irq.raise_ts
        if self.min_raise_latency is None or \
           raise_latency < self.min_raise_latency:
            self.min_raise_latency = raise_latency

        if self.max_raise_latency is None or \
           raise_latency > self.max_raise_latency:
            self.max_raise_latency = raise_latency

        self.total_raise_latency += raise_latency
        self.raise_count += 1

    def reset(self):
        super().reset()
        self.min_raise_latency = None
        self.max_raise_latency = None
        self.total_raise_latency = 0
        self.raise_count = 0



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import sp, sv


class IrqStateProvider(sp.StateProvider):
    def __init__(self, state):
        cbs = {
            'irq_handler_entry': self._process_irq_handler_entry,
            'irq_handler_exit': self._process_irq_handler_exit,
            'softirq_raise': self._process_softirq_raise,
            'softirq_entry': self._process_softirq_entry,
            'softirq_exit': self._process_softirq_exit
        }

        super().__init__(state, cbs)

    def _get_cpu(self, cpu_id):
        if cpu_id not in self._state.cpus:
            self._state.cpus[cpu_id] = sv.CPU(cpu_id)

        return self._state.cpus[cpu_id]

    # Hard IRQs
    def _process_irq_handler_entry(self, event):
        cpu = self._get_cpu(event['cpu_id'])
        irq = sv.HardIRQ.new_from_irq_handler_entry(event)
        cpu.current_hard_irq = irq

        self._state.send_notification_cb('irq_handler_entry',
                                         id=irq.id,
                                         irq_name=event['name'])

    def _process_irq_handler_exit(self, event):
        cpu = self._get_cpu(event['cpu_id'])
        if cpu.current_hard_irq is None or \
           cpu.current_hard_irq.id != event['irq']:
            cpu.current_hard_irq = None
            return

        cpu.current_hard_irq.end_ts = event.timestamp
        cpu.current_hard_irq.ret = event['ret']

        self._state.send_notification_cb('irq_handler_exit',
                                         hard_irq=cpu.current_hard_irq)
        cpu.current_hard_irq = None

    # SoftIRQs
    def _process_softirq_raise(self, event):
        cpu = self._get_cpu(event['cpu_id'])
        vec = event['vec']

        if vec not in cpu.current_softirqs:
            cpu.current_softirqs[vec] = []

        # Don't append a SoftIRQ object if one has already been raised,
        # because they are level-triggered. The only exception to this
        # is if the first SoftIRQ object already had a begin_ts which
        # means this raise was triggered after its entry, and will be
        # handled in the following softirq_entry
        if cpu.current_softirqs[vec] and \
           cpu.current_softirqs[vec][0].begin_ts is None:
            return

        irq = sv.SoftIRQ.new_from_softirq_raise(event)
        cpu.current_softirqs[vec].append(irq)

    def _process_softirq_entry(self, event):
        cpu = self._get_cpu(event['cpu_id'])
        vec = event['vec']

        if vec not in cpu.current_softirqs:
            cpu.current_softirqs[vec] = []

        if cpu.current_softirqs[vec]:
            cpu.current_softirqs[vec][0].begin_ts = event.timestamp
        else:
            # SoftIRQ entry without a corresponding raise
            irq = sv.SoftIRQ.new_from_softirq_entry(event)
            cpu.current_softirqs[vec].append(irq)

    def _process_softirq_exit(self, event):
        cpu = self._get_cpu(event['cpu_id'])
        vec = event['vec']
        # List of enqueued softirqs for the current cpu/vec
        # combination. None if vec is not found in the dictionary.
        current_softirqs = cpu.current_softirqs.get(vec)

        # Ignore the exit if either vec was not in the cpu's dict or
        # if its irq list was empty (i.e. no matching raise).
        if not current_softirqs:
            return

        current_softirqs[0].end_ts = event.timestamp
        self._state.send_notification_cb('softirq_exit',
                                         softirq=current_softirqs[0])
        del current_softirqs[0]



# The MIT License (MIT)
#
# Copyright (C) 2016 - Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import event as core_event
from functools import partial
import babeltrace as bt
import enum


class InvalidPeriodDefinition(Exception):
    pass


# period definition registry, owner of the whole tree of periods
class PeriodDefinitionRegistry:
    def __init__(self):
        self._root_period_defs = set()
        self._named_period_defs = {}
        # name to hierarchy
        self._full_period_path = {}

    def period_full_path(self, name):
        return self._full_period_path[name]

    def has_period_def(self, name):
        return name in self._named_period_defs

    def add_full_period_path(self, period_name, parent_name):
        period_path = [period_name]
        period_path_str = ""
        if parent_name is None:
            self._full_period_path[period_name] = period_name
            return

        parent = self.get_period_def(parent_name)
        while parent is not None:
            period_path.append(parent.name)
            parent = parent.parent
        period_path.reverse()
        for i in period_path:
            if len(period_path_str) == 0:
                period_path_str = i
            else:
                period_path_str = "%s/%s" % (period_path_str, i)
        self._full_period_path[period_name] = period_path_str

    def add_period_def(self, parent_name, period_name, begin_expr, end_expr,
                       begin_captures_exprs, end_captures_exprs):
        # validate unique period name (if named)
        if self.has_period_def(period_name):
            raise InvalidPeriodDefinition('Cannot redefine period "{}"'.format(
                period_name))

        # validate that parent exists if it's set
        if parent_name is not None and not self.has_period_def(parent_name):
            fmt = 'Cannot find parent period named "{}" (as parent of ' \
                  'period "{}")'
            msg = fmt.format(parent_name, period_name)
            raise InvalidPeriodDefinition(msg)

        # create period, and associate parent and children
        parent = None

        if parent_name is not None:
            parent = self.get_period_def(parent_name)

        period_def = PeriodDefinition(parent, period_name, begin_expr,
                                      end_expr, begin_captures_exprs,
                                      end_captures_exprs)

        if parent is not None:
            parent.children.add(period_def)

        # validate new period definition
        PeriodDefinitionValidator(period_def)

        if period_def.parent is None:
            self._root_period_defs.add(period_def)

        if period_def.name is not None:
            self._named_period_defs[period_def.name] = period_def

        self.add_full_period_path(period_name, parent_name)

    def get_period_def(self, name):
        return self._named_period_defs.get(name)

    @property
    def root_period_defs(self):
        for period_def in self._root_period_defs:
            yield period_def

    @property
    def named_period_defs(self):
        return self._named_period_defs

    @property
    def is_empty(self):
        return len(self._root_period_defs) == 0 and \
            len(self._named_period_defs) == 0


# definition of a period
class PeriodDefinition:
    def __init__(self, parent, name, begin_expr, end_expr,
                 begin_captures_exprs, end_captures_exprs):
        self._parent = parent
        self._children = set()
        self._name = name
        self._begin_expr = begin_expr
        self._end_expr = end_expr
        self._begin_captures_exprs = begin_captures_exprs
        self._end_captures_exprs = end_captures_exprs

    @property
    def name(self):
        return self._name

    @property
    def parent(self):
        return self._parent

    @property
    def begin_expr(self):
        return self._begin_expr

    @property
    def end_expr(self):
        return self._end_expr

    @property
    def begin_captures_exprs(self):
        return self._begin_captures_exprs

    @property
    def end_captures_exprs(self):
        return self._end_captures_exprs

    @property
    def children(self):
        return self._children


class _Expression:
    pass


class _BinaryExpression(_Expression):
    def __init__(self, lh_expr, rh_expr):
        self._lh_expr = lh_expr
        self._rh_expr = rh_expr

    @property
    def lh_expr(self):
        return self._lh_expr

    @property
    def rh_expr(self):
        return self._rh_expr


class _UnaryExpression(_Expression):
    def __init__(self, expr):
        self._expr = expr

    @property
    def expr(self):
        return self._expr


class LogicalNot(_UnaryExpression):
    def __repr__(self):
        return '!({})'.format(self.expr)


class LogicalAnd(_BinaryExpression):
    def __repr__(self):
        return '({} && {})'.format(self.lh_expr, self.rh_expr)


class LogicalOr(_BinaryExpression):
    def __repr__(self):
        return '({} || {})'.format(self.lh_expr, self.rh_expr)


class GlobEq(_BinaryExpression):
    def __init__(self, lh_expr, rh_expr):
        super().__init__(lh_expr, rh_expr)
        self._compile()

    def _compile(self):
        import fnmatch
        import re

        pattern = self.rh_expr.value
        regex = fnmatch.translate(pattern)
        self._regex = re.compile(regex)

    @property
    def regex(self):
        return self._regex

    def __repr__(self):
        return '({} =* {})'.format(self.lh_expr, self.rh_expr)


class Eq(_BinaryExpression):
    def __repr__(self):
        return '({} == {})'.format(self.lh_expr, self.rh_expr)


class Lt(_BinaryExpression):
    def __repr__(self):
        return '({} < {})'.format(self.lh_expr, self.rh_expr)


class LtEq(_BinaryExpression):
    def __repr__(self):
        return '({} <= {})'.format(self.lh_expr, self.rh_expr)


class Gt(_BinaryExpression):
    def __repr__(self):
        return '({} > {})'.format(self.lh_expr, self.rh_expr)


class GtEq(_BinaryExpression):
    def __repr__(self):
        return '({} >= {})'.format(self.lh_expr, self.rh_expr)


class Number(_Expression):
    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def __repr__(self):
        return '{}'.format(self.value)


class String(_Expression):
    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def __repr__(self):
        return '"{}"'.format(self.value)


@enum.unique
class DynScope(enum.Enum):
    AUTO = 'auto'
    TPH = '$pkt_header'
    SPC = '$pkt_ctx'
    SEH = '$header'
    SEC = '$stream_ctx'
    EC = '$ctx'
    EP = '$payload'


class _SingleChildNode(_Expression):
    def __init__(self, child):
        self._child = child

    @property
    def child(self):
        return self._child


class ParentScope(_SingleChildNode):
    def __repr__(self):
        return '$parent.{}'.format(self.child)


class BeginScope(_SingleChildNode):
    def __repr__(self):
        return '$begin.{}'.format(self.child)


class EventScope(_SingleChildNode):
    def __repr__(self):
        return '$evt.{}'.format(self.child)


class DynamicScope(_SingleChildNode):
    def __init__(self, dyn_scope, child):
        super().__init__(child)
        self._dyn_scope = dyn_scope

    @property
    def dyn_scope(self):
        return self._dyn_scope

    def __repr__(self):
        if self._dyn_scope == DynScope.AUTO:
            return repr(self.child)

        return '{}.{}'.format(self.dyn_scope.value, self.child)


class EventFieldName(_Expression):
    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

    def __repr__(self):
        return self._name


class EventName(_Expression):
    def __repr__(self):
        return '$name'


class IllegalExpression(Exception):
    pass


class PeriodDefinitionValidator:
    def __init__(self, period_def):
        self._period_def = period_def
        self._validate_expr_cbs = {
            LogicalNot: self._validate_unary_expr,
            LogicalAnd: self._validate_binary_expr,
            LogicalOr: self._validate_binary_expr,
            GlobEq: self._validate_comp,
            Eq: self._validate_comp,
            Lt: self._validate_comp,
            LtEq: self._validate_comp,
            Gt: self._validate_comp,
            GtEq: self._validate_comp,
            ParentScope: self._validate_parent_scope,
        }
        self._validate_expr(period_def.begin_expr)
        self._validate_expr(period_def.end_expr)

    def _validate_unary_expr(self, not_expr):
        self._validate_expr(not_expr.expr)

    def _validate_binary_expr(self, and_expr):
        self._validate_expr(and_expr.lh_expr)
        self._validate_expr(and_expr.rh_expr)

    def _validate_parent_scope(self, scope):
        if self._period_def.parent is None:
            raise IllegalExpression('Cannot refer to parent context without '
                                    'a named parent period')

        if type(scope.child) is not BeginScope:
            raise IllegalExpression('Must refer to the begin context in a '
                                    'parent context')

        self._validate_expr(scope.child)

    def _validate_comp(self, comp_expr):
        self._validate_expr(comp_expr.lh_expr)
        self._validate_expr(comp_expr.rh_expr)

    def _validate_expr(self, expr):
        if type(expr) in self._validate_expr_cbs:
            self._validate_expr_cbs[type(expr)](expr)


class _MatchContext:
    def __init__(self, evt, begin_evt, parent_begin_evt):
        self._evt = evt
        self._begin_evt = begin_evt
        self._parent_begin_evt = parent_begin_evt

    @property
    def evt(self):
        return self._evt

    @property
    def begin_evt(self):
        return self._begin_evt

    @property
    def parent_begin_evt(self):
        return self._parent_begin_evt


_DYN_SCOPE_TO_BT_CTF_SCOPE = {
    DynScope.TPH: bt.CTFScope.TRACE_PACKET_HEADER,
    DynScope.SPC: bt.CTFScope.STREAM_PACKET_CONTEXT,
    DynScope.SEH: bt.CTFScope.STREAM_EVENT_HEADER,
    DynScope.SEC: bt.CTFScope.STREAM_EVENT_CONTEXT,
    DynScope.EC: bt.CTFScope.EVENT_CONTEXT,
    DynScope.EP: bt.CTFScope.EVENT_FIELDS,
}


def _resolve_event_expr(event, expr):
    # event not found
    if event is None:
        return

    # event name
    if type(expr.child) is EventName:
        return event.name

    # default, automatic dynamic scope
    dyn_scope = DynScope.AUTO

    if type(expr.child) is DynamicScope:
        # select specific dynamic scope
        expr = expr.child
        dyn_scope = expr.dyn_scope

    if type(expr.child) is EventFieldName:
        expr = expr.child

        if dyn_scope == DynScope.AUTO:
            # automatic dynamic scope
            if expr.name in event:
                return event[expr.name]

            # event field not found
            return

        # specific dynamic scope
        bt_ctf_scope = _DYN_SCOPE_TO_BT_CTF_SCOPE[dyn_scope]

        return event.field_with_scope(expr.name, bt_ctf_scope)

    assert(False)


# This exquisite function takes an expression and resolves it to
# an actual value (Python's number/string) considering the current
# matching context.
def _resolve_expr(expr, match_context):
    if type(expr) is ParentScope:
        begin_scope = expr.child
        event_scope = begin_scope.child

        return _resolve_event_expr(match_context.parent_begin_evt, event_scope)

    if type(expr) is BeginScope:
        # event in the begin context
        event_scope = expr.child

        return _resolve_event_expr(match_context.begin_evt, event_scope)

    if type(expr) is EventScope:
        # current event
        return _resolve_event_expr(match_context.evt, expr)

    if type(expr) is Number:
        return expr.value

    if type(expr) is String:
        return expr.value

    assert(False)


class _Matcher:
    def __init__(self, expr, match_context):
        self._match_context = match_context
        self._expr_matchers = {
            LogicalAnd: self._and_expr_matches,
            LogicalOr: self._or_expr_matches,
            LogicalNot: self._not_expr_matches,
            GlobEq: self._glob_eq_expr_matches,
            Eq: partial(self._comp_expr_matches, lambda lh, rh: lh == rh),
            Lt: partial(self._comp_expr_matches, lambda lh, rh: lh < rh),
            LtEq: partial(self._comp_expr_matches, lambda lh, rh: lh <= rh),
            Gt: partial(self._comp_expr_matches, lambda lh, rh: lh > rh),
            GtEq: partial(self._comp_expr_matches, lambda lh, rh: lh >= rh),
        }
        self._matches = self._expr_matches(expr)

    def _and_expr_matches(self, expr):
        return (self._expr_matches(expr.lh_expr) and
                self._expr_matches(expr.rh_expr))

    def _or_expr_matches(self, expr):
        return (self._expr_matches(expr.lh_expr) or
                self._expr_matches(expr.rh_expr))

    def _not_expr_matches(self, expr):
        return not self._expr_matches(expr.expr)

    def _glob_eq_expr_matches(self, expr):
        def compfn(lh, rh):
            return bool(expr.regex.match(lh))

        return self._comp_expr_matches(compfn, expr)

    def _comp_expr_matches(self, compfn, expr):
        lh_value = _resolve_expr(expr.lh_expr, self._match_context)
        rh_value = _resolve_expr(expr.rh_expr, self._match_context)

        # make sure both sides are found
        if lh_value is None or rh_value is None:
            return False

        # cast RHS to int if LHS is an int
        if type(lh_value) is int and type(rh_value) is float:
            rh_value = int(rh_value)

        # compare types first
        if type(lh_value) is not type(rh_value):
            return False

        # compare field to a literal value
        return compfn(lh_value, rh_value)

    def _expr_matches(self, expr):
        return self._expr_matchers[type(expr)](expr)

    @property
    def matches(self):
        return self._matches


def _expr_matches(expr, match_context):
    return _Matcher(expr, match_context).matches


def create_conjunction_from_exprs(exprs):
    if len(exprs) == 0:
        return

    cur_expr = exprs[0]

    for expr in exprs[1:]:
        cur_expr = LogicalAnd(cur_expr, expr)

    return cur_expr


def create_disjunction_from_exprs(exprs):
    if len(exprs) == 0:
        return

    cur_expr = exprs[0]

    for expr in exprs[1:]:
        cur_expr = LogicalOr(cur_expr, expr)

    return cur_expr


@enum.unique
class PeriodEngineCallbackType(enum.Enum):
    PERIOD_BEGIN = 1
    PERIOD_END = 2


class Period:
    def __init__(self, definition, parent, begin_evt, begin_captures):
        begin_evt_copy = core_event.Event(begin_evt)
        self._begin_evt = begin_evt_copy
        self._end_evt = None
        self._completed = False
        self._definition = definition
        self._parent = parent
        self._children = set()
        self._begin_captures = begin_captures
        self._end_captures = {}

    @property
    def begin_evt(self):
        return self._begin_evt

    @property
    def end_evt(self):
        return self._end_evt

    @end_evt.setter
    def end_evt(self, evt):
        self._end_evt = evt

    @property
    def definition(self):
        return self._definition

    @property
    def parent(self):
        return self._parent

    @property
    def children(self):
        return self._children

    @property
    def completed(self):
        return self._completed

    @completed.setter
    def completed(self, value):
        self._completed = value

    @property
    def begin_captures(self):
        return self._begin_captures

    @property
    def end_captures(self):
        return self._end_captures


class PeriodEngine:
    def __init__(self, registry, cbs):
        self._registry = registry
        self._cbs = cbs
        self._root_periods = set()

    def _cb_period_end(self, period):
        self._cbs[PeriodEngineCallbackType.PERIOD_END](period)

    def _cb_period_begin(self, period):
        self._cbs[PeriodEngineCallbackType.PERIOD_BEGIN](period)

    def _create_period(self, definition, parent, begin_evt, begin_captures):
        return Period(definition, parent, begin_evt, begin_captures)

    def _get_captures(self, captures_exprs, match_context):
        captures = {}

        for name, capture_expr in captures_exprs.items():
            captures[name] = _resolve_expr(capture_expr, match_context)

        return captures

    def _process_event_add_periods(self, parent_period,
                                   child_periods, child_period_defs, evt):
        periods_to_add = set()

        for child_period_def in child_period_defs:
            match_context = self._create_begin_match_context(parent_period,
                                                             evt)

            if _expr_matches(child_period_def.begin_expr, match_context):
                # match! add period
                captures = self._get_captures(
                    child_period_def.begin_captures_exprs,
                    match_context)
                period = self._create_period(child_period_def,
                                             parent_period, evt, captures)
                periods_to_add.add(period)

        # safe to add child periods now, outside the iteration
        for period_to_add in periods_to_add:
            self._cb_period_begin(period_to_add)
            child_periods.add(period_to_add)

        for child_period in child_periods:
            self._process_event_add_periods(child_period,
                                            child_period.children,
                                            child_period.definition.children,
                                            evt)

    def _process_event_begin(self, evt):
        defs = self._registry.root_period_defs
        self._process_event_add_periods(None, self._root_periods, defs, evt)

    def _create_begin_match_context(self, parent_period, evt):
        parent_begin_evt = None

        if parent_period is not None:
            parent_begin_evt = parent_period.begin_evt

        return _MatchContext(evt, evt, parent_begin_evt)

    def _create_end_match_context(self, period, evt):
        parent_begin_evt = None

        if period.parent is not None:
            parent_begin_evt = period.parent.begin_evt

        return _MatchContext(evt, period.begin_evt, parent_begin_evt)

    def _process_event_remove_period(self, child_periods, evt):
        for child_period in child_periods:
            self._process_event_remove_period(child_period.children, evt)

        child_periods_to_remove = set()

        for child_period in child_periods:
            match_context = self._create_end_match_context(child_period, evt)

            if _expr_matches(child_period.definition.end_expr, match_context):
                # set period's end captures
                end_captures_exprs = \
                    child_period.definition.end_captures_exprs
                captures = self._get_captures(end_captures_exprs,
                                              match_context)
                child_period._end_captures = captures

                # mark as to be removed
                child_periods_to_remove.add(child_period)

        # safe to remove child periods now, outside the iteration
        for child_period_to_remove in child_periods_to_remove:
            # set period's ending event and completed property
            child_period_to_remove.end_evt = evt
            child_period_to_remove.completed = True

            # also remove its own remaining child periods
            self._remove_periods(child_period_to_remove.children, evt)

            # call end of period user callback (this period matched)
            self._cb_period_end(child_period_to_remove)

            # remove period from set
            child_periods.remove(child_period_to_remove)

    def _process_event_end(self, evt):
        self._process_event_remove_period(self._root_periods, evt)

    def process_event(self, evt):
        self._process_event_end(evt)
        self._process_event_begin(evt)

    def _remove_periods(self, child_periods, evt):
        for child_period in child_periods:
            self._remove_periods(child_period.children, evt)

        # safe to remove child periods now, outside the iteration
        for child_period in child_periods:
            # set period's ending event and completed property
            child_period.end_evt = evt
            child_period.completed = False

            # call end of period user callback
            self._cb_period_end(child_period)

        child_periods.clear()

    def remove_all_periods(self):
        self._remove_periods(self._root_periods, None)

    @property
    def root_periods(self):
        return self._root_periods



# The MIT License (MIT)
#
# Copyright (C) 2016 - Julien Desfossez <jdesfossez@efficios.com>
#                      Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .analysis_test import AnalysisTest


class IoTest(AnalysisTest):
    def write_trace(self):
        # app (99) is known at statedump
        self.trace_writer.write_lttng_statedump_process_state(
            1000, 0, 99, 99, 99, 99, 98, 98, 'app', 0, 5, 0, 5, 0)
        # app2 (100) unknown at statedump has testfile, FD 3 defined at
        # statedump
        self.trace_writer.write_lttng_statedump_file_descriptor(
            1001, 0, 100, 3, 0, 0, 'testfile')
        # app write 10 bytes to FD 4
        self.trace_writer.write_sched_switch(1002, 0, 'swapper/0', 0, 'app',
                                             99)
        self.trace_writer.write_syscall_write(1004, 0, 1, 4, 0xabcd, 10, 10)
        # app2 reads 100 bytes in FD 3
        self.trace_writer.write_sched_switch(1006, 0, 'app', 99, 'app2', 100)
        self.trace_writer.write_syscall_read(1008, 0, 1, 3, 0xcafe, 100, 100)
        # app3 and its FD 3 are completely unknown at statedump, tries to read
        # 100 bytes from FD 3 but only gets 42
        self.trace_writer.write_sched_switch(1010, 0, 'app2', 100, 'app3', 101)
        self.trace_writer.write_syscall_read(1012, 0, 1, 3, 0xcafe, 100, 42)
        # block write
        self.trace_writer.write_block_rq_issue(1015, 0, 264241152, 33, 10, 40,
                                               99, 0, 0, '', 'app')
        self.trace_writer.write_block_rq_complete(1016, 0, 264241152, 33, 10,
                                                  0, 0, 0, '')
        # block read
        self.trace_writer.write_block_rq_issue(1017, 0, 8388608, 33, 20, 90,
                                               101, 1, 0, '', 'app3')
        self.trace_writer.write_block_rq_complete(1018, 0, 8388608, 33, 20, 0,
                                                  1, 0, '')
        # net xmit
        self.trace_writer.write_net_dev_xmit(1020, 2, 0xff, 32, 100, 'wlan0')
        # net receive
        self.trace_writer.write_netif_receive_skb(1021, 1, 0xff, 100, 'wlan1')
        self.trace_writer.write_netif_receive_skb(1022, 1, 0xff, 200, 'wlan0')
        # syscall open
        self.trace_writer.write_syscall_open(1023, 0, 1, 'test/open/file', 0,
                                             0, 42)
        self.trace_writer.flush()

    def test_iousagetop(self):
        test_name = 'iousagetop'
        expected = self.get_expected_output(test_name)
        result = self.get_cmd_output('lttng-iousagetop',
                                     options='--no-intersection')

        self._assertMultiLineEqual(result, expected, test_name)

    def test_iolatencytop(self):
        test_name = 'iolatencytop'
        expected = self.get_expected_output(test_name)
        result = self.get_cmd_output('lttng-iolatencytop',
                                     options='--no-intersection')

        self._assertMultiLineEqual(result, expected, test_name)



# The MIT License (MIT)
#
# Copyright (C) 2016 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time
import datetime
import subprocess
import sys
from .version_utils import Version
from .time_utils import NSEC_PER_SEC


BT_INTERSECT_VERSION = Version(1, 4, 0)


def is_multi_day_trace_collection_bt_1_3_2(collection, handles=None):
    """is_multi_day_trace_collection for BT < 1.3.3.

    Args:
        collection (TraceCollection): a babeltrace TraceCollection
        instance.

        handles (TraceHandle): a babeltrace TraceHandle instance.

    Returns:
        True if the trace collection spans more than one day,
        False otherwise.
    """

    time_begin = None

    for handle in handles.values():
        if time_begin is None:
            time_begin = time.localtime(handle.timestamp_begin / NSEC_PER_SEC)
            year_begin = time_begin.tm_year
            month_begin = time_begin.tm_mon
            day_begin = time_begin.tm_mday

        time_end = time.localtime(handle.timestamp_end / NSEC_PER_SEC)
        year_end = time_end.tm_year
        month_end = time_end.tm_mon
        day_end = time_end.tm_mday

        if year_begin != year_end:
            return True
        elif month_begin != month_end:
            return True
        elif day_begin != day_end:
            return True

    return False


def is_multi_day_trace_collection(collection, handles=None):
    """Check whether a trace collection spans more than one day.

    Args:
        collection (TraceCollection): a babeltrace TraceCollection
        instance.
        handles (TraceHandle): a babeltrace TraceHandle instance.

    Returns:
        True if the trace collection spans more than one day,
        False otherwise.
    """

    # Circumvent a bug in Babeltrace < 1.3.3
    if collection.timestamp_begin is None or \
            collection.timestamp_end is None:
                return is_multi_day_trace_collection_bt_1_3_2(collection,
                                                              handles)

    date_begin = datetime.date.fromtimestamp(
        collection.timestamp_begin // NSEC_PER_SEC
    )
    date_end = datetime.date.fromtimestamp(
        collection.timestamp_end // NSEC_PER_SEC
    )

    return date_begin != date_end


def get_trace_collection_date(collection, handles=None):
    """Get a trace collection's date.

    Args:
        collection (TraceCollection): a babeltrace TraceCollection
        instance.

        handles (TraceHandle): a babeltrace TraceHandle instance.

    Returns:
        A datetime.date object corresponding to the date at which the
        trace collection was recorded.

        handles (TraceHandle): a babeltrace TraceHandle instance.

    Raises:
        ValueError: if the trace collection spans more than one day.
    """
    if is_multi_day_trace_collection(collection, handles):
        raise ValueError('Trace collection spans multiple days')

    trace_date = datetime.date.fromtimestamp(
        collection.timestamp_begin // NSEC_PER_SEC
    )

    return trace_date


def get_syscall_name(event):
    """Get the name of a syscall from an event.

    Args:
        event (Event): an instance of a babeltrace Event for a syscall
        entry.

    Returns:
        The name of the syscall, stripped of any superfluous prefix.

    Raises:
        ValueError: if the event is not a syscall event.
    """
    name = event.name

    if name.startswith('sys_'):
        return name[4:]
    elif name.startswith('syscall_entry_'):
        return name[14:]
    else:
        raise ValueError('Not a syscall event')


def read_babeltrace_version():
    try:
        output = subprocess.check_output('babeltrace')
    except subprocess.CalledProcessError:
        raise ValueError('Could not run babeltrace to verify version')

    output = output.decode(sys.stdout.encoding)
    first_line = output.splitlines()[0]
    version_string = first_line.split()[-1]

    return Version.new_from_string(version_string)


def check_field_exists(handles, ev_name, field_name):
    """Validate that a field exists in the metadata.

    Args:
        handles (TraceHandle): an array of babeltrace TraceHandle instance.

        ev_name (String): the event name in which the field must exist.

        field_name (String): the field that we are looking for.

    Returns:
        True if the field is found in the event, False if the field is not
        found in the event, or if the event is not found.
    """
    for handle in handles.values():
        for event in handle.events:
            if event.name == ev_name:
                for field in event.fields:
                    if field.name == field_name:
                        return True
    return False


def check_event_exists(handles, name):
    """Validate that an event exists in the metadata.

    Args:
        handles (TraceHandle): an array of babeltrace TraceHandle instance.

        name (String): the event name in which the field must exist.

    Returns:
        True if the event is found in the metadata, False otherwise.
    """
    for handle in handles.values():
        for event in handle.events:
            if event.name == name:
                return True
    return False



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import sp, sv


class BlockStateProvider(sp.StateProvider):
    def __init__(self, state):
        cbs = {
            'block_rq_complete': self._process_block_rq_complete,
            'block_rq_issue': self._process_block_rq_issue,
            'block_bio_remap': self._process_block_bio_remap,
            'block_bio_backmerge': self._process_block_bio_backmerge,
        }

        super().__init__(state, cbs)
        self._remap_requests = []

    def _process_block_bio_remap(self, event):
        dev = event['dev']
        sector = event['sector']
        old_dev = event['old_dev']
        old_sector = event['old_sector']

        for req in self._remap_requests:
            if req.dev == old_dev and req.sector == old_sector:
                req.dev = dev
                req.sector = sector
                return

        req = sv.BlockRemapRequest(dev, sector, old_dev, old_sector)
        self._remap_requests.append(req)

    # For backmerge requests, just remove the request from the
    # _remap_requests queue, because we rely later on the nr_sector
    # which has all the info we need
    def _process_block_bio_backmerge(self, event):
        dev = event['dev']
        sector = event['sector']
        for remap_req in self._remap_requests:
            if remap_req.dev == dev and remap_req.sector == sector:
                self._remap_requests.remove(remap_req)

    def _process_block_rq_issue(self, event):
        dev = event['dev']
        sector = event['sector']
        nr_sector = event['nr_sector']

        if nr_sector == 0:
            return

        req = sv.BlockIORequest.new_from_rq_issue(event)

        for remap_req in self._remap_requests:
            if remap_req.dev == dev and remap_req.sector == sector:
                dev = remap_req.old_dev
                break

        if dev not in self._state.disks:
            self._state.disks[dev] = sv.Disk(dev)

        self._state.disks[dev].pending_requests[sector] = req

    def _process_block_rq_complete(self, event):
        dev = event['dev']
        sector = event['sector']
        nr_sector = event['nr_sector']

        if nr_sector == 0:
            return

        for remap_req in self._remap_requests:
            if remap_req.dev == dev and remap_req.sector == sector:
                dev = remap_req.old_dev
                self._remap_requests.remove(remap_req)
                break

        if dev not in self._state.disks:
            self._state.disks[dev] = sv.Disk(dev)

        disk = self._state.disks[dev]

        # Ignore rq_complete without matching rq_issue
        if sector not in disk.pending_requests:
            return

        req = disk.pending_requests[sector]
        # Ignore rq_complete if nr_sector does not match rq_issue's
        if req.nr_sector != nr_sector:
            return

        req.update_from_rq_complete(event)
        if req.tid in self._state.tids.keys():
            proc = self._state.tids[req.tid]
        else:
            proc = None
        self._state.send_notification_cb('block_rq_complete', req=req,
                                         proc=proc, cpu_id=event['cpu_id'],
                                         disk=disk)
        del disk.pending_requests[sector]



# The MIT License (MIT)
#
# Copyright (C) 2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import stats
from .analysis import Analysis, PeriodData


class _PeriodData(PeriodData):
    def __init__(self):
        self.tids = {}


class Memtop(Analysis):
    def __init__(self, state, conf):
        notification_cbs = {
            'tid_page_alloc': self._process_tid_page_alloc,
            'tid_page_free': self._process_tid_page_free
        }
        super().__init__(state, conf, notification_cbs)

    def _create_period_data(self):
        return _PeriodData()

    def _process_tid_page_alloc(self, period_data, **kwargs):
        cpu_id = kwargs['cpu_id']
        proc = kwargs['proc']

        if not self._filter_process(proc):
            return
        if not self._filter_cpu(cpu_id):
            return

        tid = proc.tid
        if tid not in period_data.tids:
            period_data.tids[tid] = ProcessMemStats.new_from_process(proc)

        period_data.tids[tid].allocated_pages += 1

    def _process_tid_page_free(self, period_data, **kwargs):
        cpu_id = kwargs['cpu_id']
        proc = kwargs['proc']

        if not self._filter_process(proc):
            return
        if not self._filter_cpu(cpu_id):
            return

        tid = proc.tid
        if tid not in period_data.tids:
            period_data.tids[tid] = ProcessMemStats.new_from_process(proc)

        period_data.tids[tid].freed_pages += 1


class ProcessMemStats(stats.Process):
    def __init__(self, pid, tid, comm):
        super().__init__(pid, tid, comm)

        self.allocated_pages = 0
        self.freed_pages = 0

    def reset(self):
        self.allocated_pages = 0
        self.freed_pages = 0



# The MIT License (MIT)
#
# Copyright (C) 2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
from functools import total_ordering


@total_ordering
class Version:
    def __init__(self, major, minor, patch, extra=None):
        self.major = major
        self.minor = minor
        self.patch = patch
        self.extra = extra

    def __lt__(self, other):
        if self.major < other.major:
            return True
        if self.major > other.major:
            return False

        if self.minor < other.minor:
            return True
        if self.minor > other.minor:
            return False

        return self.patch < other.patch

    def __eq__(self, other):
        return (
            self.major == other.major and
            self.minor == other.minor and
            self.patch == other.patch
        )

    def __repr__(self):
        version_str = '{}.{}.{}'.format(self.major, self.minor, self.patch)
        if self.extra:
            version_str += self.extra

        return version_str

    @classmethod
    def new_from_string(cls, string):
        version_match = re.match(r'(\d+)\.(\d+)\.(\d+)(.*)', string)

        if version_match is None:
            major = minor = patch = 0
            extra = '+unknown'
        else:
            major = int(version_match.group(1))
            minor = int(version_match.group(2))
            patch = int(version_match.group(3))
            extra = version_match.group(4)

        return cls(major, minor, patch, extra)



# The MIT License (MIT)
#
# Copyright (C) 2016 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math
import socket
import struct
import time
from .time_utils import NSEC_PER_SEC


def format_size(size, binary_prefix=True):
    """Convert an integral number of bytes to a human-readable string.

    Args:
        size (int): a non-negative number of bytes.

        binary_prefix (bool, optional): whether to use binary units
        prefixes, over SI prefixes (default: True).

    Returns:
        The formatted string comprised of the size and units.

    Raises:
        ValueError: if size < 0.
    """
    if size < 0:
        raise ValueError('Cannot format negative size')

    if binary_prefix:
        base = 1024
        units = ['  B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB']
    else:
        base = 1000
        units = [' B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']

    if size == 0:
        exponent = 0
    else:
        exponent = int(math.log(size, base))
        if exponent >= len(units):
            # Don't try and use a unit above YiB/YB
            exponent = len(units) - 1

        size /= base ** exponent

    unit = units[exponent]

    if exponent == 0:
        # Don't display fractions of a byte
        format_str = '{:0.0f} {}'
    else:
        format_str = '{:0.2f} {}'

    return format_str.format(size, unit)


def format_prio_list(prio_list):
    """Format a list of prios into a string of unique prios with count.

    Args:
        prio_list (list): a list of PrioEvent objects.

    Returns:
        The formatted string containing the unique priorities and
        their count if they occurred more than once.
    """
    prio_count = {}
    prio_str = None

    for prio_event in prio_list:
        prio = prio_event.prio
        if prio not in prio_count:
            prio_count[prio] = 0

        prio_count[prio] += 1

    for prio in sorted(prio_count.keys()):
        count = prio_count[prio]
        if count > 1:
            count_str = ' ({} times)'.format(count)
        else:
            count_str = ''

        if prio_str is None:
            prio_str = '[{}{}'.format(prio, count_str)
        else:
            prio_str += ', {}{}'.format(prio, count_str)

    if prio_str is None:
        prio_str = '[]'
    else:
        prio_str += ']'

    return prio_str


def format_timestamp(timestamp, print_date=False, gmt=False):
    """Format a timestamp into a human-readable date string

    Args:
        timestamp (int): nanoseconds since epoch.

        print_date (bool, optional): flag indicating whether to print
        the full date or just the time of day (default: False).

        gmt (bool, optional): flag indicating whether the timestamp is
        in the local timezone or gmt (default: False).

    Returns:
        The formatted date string, containing either the full date or
        just the time of day.
    """
    date_fmt = '{:04}-{:02}-{:02} '
    time_fmt = '{:02}:{:02}:{:02}.{:09}'

    if gmt:
        date = time.gmtime(timestamp // NSEC_PER_SEC)
    else:
        date = time.localtime(timestamp // NSEC_PER_SEC)

    formatted_ts = time_fmt.format(
        date.tm_hour, date.tm_min, date.tm_sec,
        timestamp % NSEC_PER_SEC
    )

    if print_date:
        date_str = date_fmt.format(date.tm_year, date.tm_mon, date.tm_mday)
        formatted_ts = date_str + formatted_ts

    return formatted_ts


def format_time_range(begin_ts, end_ts, print_date=False, gmt=False):
    """Format a pair of timestamps into a human-readable date string.

    Args:
        begin_ts (int): nanoseconds since epoch to beginning of
        time range.

        end_ts (int): nanoseconds since epoch to end of time range.

        print_date (bool, optional): flag indicating whether to print
        the full date or just the time of day (default: False).

        gmt (bool, optional): flag indicating whether the timestamp is
        in the local timezone or gmt (default: False).

    Returns:
        The formatted dates string, containing either the full date or
        just the time of day, enclosed within square brackets and
        delimited by a comma.
    """
    time_range_fmt = '[{}, {}]'

    begin_str = format_timestamp(begin_ts, print_date, gmt)
    end_str = format_timestamp(end_ts, print_date, gmt)

    return time_range_fmt.format(begin_str, end_str)


def format_ipv4(ip, port=None):
    """Format an ipv4 address into a human-readable string.

    Args:
        ip (varies): the ip address as extracted in an LTTng event.
        Either an integer or a list of integers, depending on the
        tracer version.

        port (int, optional): the port number associated with the
        address.

    Returns:
        The formatted string containing the ipv4 address and, optionally,
        the port number.

    """
    # depending on the version of lttng-modules, the v4addr is an
    # integer (< 2.6) or sequence (>= 2.6)
    try:
        ip_str = '{}.{}.{}.{}'.format(ip[0], ip[1], ip[2], ip[3])
    except TypeError:
        # The format string '!I' tells pack to interpret ip as a
        # packed structure of network-endian 32-bit unsigned integers,
        # which inet_ntoa can then convert into the formatted string
        ip_str = socket.inet_ntoa(struct.pack('!I', ip))

    if port is not None:
        ip_str += ':{}'.format(port)

    return ip_str



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import math
import operator
import statistics
import collections
from . import mi, termgraph
from ..core import sched
from .command import Command
from ..common import format_utils


_SchedStats = collections.namedtuple('_SchedStats', [
    'count',
    'min',
    'max',
    'stdev',
    'total',
])


class SchedAnalysisCommand(Command):
    _DESC = """The sched command."""
    _ANALYSIS_CLASS = sched.SchedAnalysis
    _MI_TITLE = 'Scheduling latencies analysis'
    _MI_DESCRIPTION = \
        'Scheduling latencies frequency distribution, statistics, top, and log'
    _MI_TAGS = [mi.Tags.SCHED, mi.Tags.STATS, mi.Tags.FREQ, mi.Tags.TOP,
                mi.Tags.LOG]
    _MI_TABLE_CLASS_LOG = 'log'
    _MI_TABLE_CLASS_TOP = 'top'
    _MI_TABLE_CLASS_TOTAL_STATS = 'total_stats'
    _MI_TABLE_CLASS_PER_TID_STATS = 'per_tid_stats'
    _MI_TABLE_CLASS_PER_PRIO_STATS = 'per_prio_stats'
    _MI_TABLE_CLASS_FREQ = 'freq'
    # _MI_TABLE_CLASS_SUMMARY = 'summary'
    _MI_TABLE_CLASSES = [
        (
            _MI_TABLE_CLASS_LOG,
            'Scheduling log', [
                ('wakeup_ts', 'Wakeup timestamp', mi.Timestamp),
                ('switch_ts', 'Switch timestamp', mi.Timestamp),
                ('latency', 'Scheduling latency', mi.Duration),
                ('prio', 'Priority', mi.Number),
                ('target_cpu', 'Target CPU', mi.Cpu),
                ('wakee_proc', 'Wakee process', mi.Process),
                ('waker_proc', 'Waker process', mi.Process),
            ]
        ),
        (
            _MI_TABLE_CLASS_TOP,
            'Scheduling top', [
                ('wakeup_ts', 'Wakeup timestamp', mi.Timestamp),
                ('switch_ts', 'Switch timestamp', mi.Timestamp),
                ('latency', 'Scheduling latency', mi.Duration),
                ('prio', 'Priority', mi.Number),
                ('target_cpu', 'Target CPU', mi.Cpu),
                ('wakee_proc', 'Wakee process', mi.Process),
                ('waker_proc', 'Waker process', mi.Process),
            ]
        ),
        (
            _MI_TABLE_CLASS_TOTAL_STATS,
            'Scheduling latency stats (total)', [
                ('count', 'Scheduling count', mi.Number, 'schedulings'),
                ('min_latency', 'Minimum latency', mi.Duration),
                ('avg_latency', 'Average latency', mi.Duration),
                ('max_latency', 'Maximum latency', mi.Duration),
                ('stdev_latency', 'Scheduling latency standard deviation',
                 mi.Duration),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_TID_STATS,
            'Scheduling latency stats (per-TID)', [
                ('process', 'Wakee process', mi.Process),
                ('count', 'Scheduling count', mi.Number, 'schedulings'),
                ('min_latency', 'Minimum latency', mi.Duration),
                ('avg_latency', 'Average latency', mi.Duration),
                ('max_latency', 'Maximum latency', mi.Duration),
                ('stdev_latency', 'Scheduling latency standard deviation',
                 mi.Duration),
                ('prio_list', 'Chronological priorities', mi.String),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_PRIO_STATS,
            'Scheduling latency stats (per-prio)', [
                ('prio', 'Priority', mi.Number),
                ('count', 'Scheduling count', mi.Number, 'schedulings'),
                ('min_latency', 'Minimum latency', mi.Duration),
                ('avg_latency', 'Average latency', mi.Duration),
                ('max_latency', 'Maximum latency', mi.Duration),
                ('stdev_latency', 'Scheduling latency standard deviation',
                 mi.Duration),
            ]
        ),
        (
            _MI_TABLE_CLASS_FREQ,
            'Scheduling latency frequency distribution', [
                ('duration_lower', 'Duration (lower bound)', mi.Duration),
                ('duration_upper', 'Duration (upper bound)', mi.Duration),
                ('count', 'Scheduling count', mi.Number, 'schedulings'),
            ]
        ),
    ]

    def _analysis_tick(self, period_data, end_ns):
        if period_data is None:
            return

        begin_ns = period_data.period.begin_evt.timestamp
        log_table = None
        top_table = None
        total_stats_table = None
        per_tid_stats_table = None
        per_prio_stats_table = None
        total_freq_tables = None
        per_tid_freq_tables = None
        per_prio_freq_tables = None

        if self._args.log:
            log_table = self._get_log_result_table(period_data,
                                                   begin_ns, end_ns)

        if self._args.top:
            top_table = self._get_top_result_table(period_data,
                                                   begin_ns, end_ns)

        if self._args.stats:
            if self._args.total:
                total_stats_table = self._get_total_stats_result_table(
                    period_data, begin_ns, end_ns)

            if self._args.per_tid:
                per_tid_stats_table = self._get_per_tid_stats_result_table(
                    period_data, begin_ns, end_ns)

            if self._args.per_prio:
                per_prio_stats_table = self._get_per_prio_stats_result_table(
                    period_data, begin_ns, end_ns)

        if self._args.freq:
            if self._args.total:
                total_freq_tables = self._get_total_freq_result_tables(
                    period_data, begin_ns, end_ns)

            if self._args.per_tid:
                per_tid_freq_tables = self._get_per_tid_freq_result_tables(
                    period_data, begin_ns, end_ns)

            if self._args.per_prio:
                per_prio_freq_tables = self._get_per_prio_freq_result_tables(
                    period_data, begin_ns, end_ns)

        if self._mi_mode:
            if log_table:
                self._mi_append_result_table(log_table)

            if top_table:
                self._mi_append_result_table(top_table)

            if total_stats_table and total_stats_table.rows:
                self._mi_append_result_table(total_stats_table)

            if per_tid_stats_table and per_tid_stats_table.rows:
                self._mi_append_result_table(per_tid_stats_table)

            if per_prio_stats_table and per_prio_stats_table.rows:
                self._mi_append_result_table(per_prio_stats_table)

            if self._args.freq:
                if total_freq_tables:
                    self._mi_append_result_tables(total_freq_tables)

                if per_tid_freq_tables:
                    if self._args.freq_series:
                        per_tid_freq_tables = [
                            self._get_per_tid_freq_series_table(
                                per_tid_freq_tables)
                        ]
                    self._mi_append_result_tables(per_tid_freq_tables)

                if per_prio_freq_tables:
                    if self._args.freq_series:
                        per_prio_freq_tables = [
                            self._get_per_prio_freq_series_table(
                                per_prio_freq_tables)
                        ]
                    self._mi_append_result_tables(per_prio_freq_tables)
        else:
            self._print_date(begin_ns, end_ns)

            if self._args.stats:
                if total_stats_table:
                    self._print_total_stats(total_stats_table)
                if per_tid_stats_table:
                    self._print_per_tid_stats(per_tid_stats_table)
                if per_prio_stats_table:
                    self._print_per_prio_stats(per_prio_stats_table)

            if self._args.freq:
                if total_freq_tables:
                    self._print_freq(total_freq_tables)
                if per_tid_freq_tables:
                    self._print_freq(per_tid_freq_tables)
                if per_prio_freq_tables:
                    self._print_freq(per_prio_freq_tables)

            if log_table:
                self._print_sched_events(log_table)

            if top_table:
                self._print_sched_events(top_table)

    def _get_total_sched_lists_stats(self, period_data):
        total_list = period_data.sched_list
        stdev = self._compute_sched_latency_stdev(total_list)
        total_stats = _SchedStats(
            count=self._analysis.count(period_data),
            min=period_data.min_latency,
            max=period_data.max_latency,
            stdev=stdev,
            total=period_data.total_latency
        )

        return [total_list], total_stats

    def _get_tid_sched_lists_stats(self, period_data):
        tid_sched_lists = {}
        tid_stats = {}

        for sched_event in period_data.sched_list:
            tid = sched_event.wakee_proc.tid
            if tid not in tid_sched_lists:
                tid_sched_lists[tid] = []

            tid_sched_lists[tid].append(sched_event)

        for tid in tid_sched_lists:
            sched_list = tid_sched_lists[tid]

            if not sched_list:
                continue

            stdev = self._compute_sched_latency_stdev(sched_list)
            latencies = [sched.latency for sched in sched_list]
            count = len(latencies)
            min_latency = min(latencies)
            max_latency = max(latencies)
            total_latency = sum(latencies)

            tid_stats[tid] = _SchedStats(
                count=count,
                min=min_latency,
                max=max_latency,
                stdev=stdev,
                total=total_latency,
            )

        return tid_sched_lists, tid_stats

    def _get_prio_sched_lists_stats(self, period_data):
        prio_sched_lists = {}
        prio_stats = {}

        for sched_event in period_data.sched_list:
            if sched_event.prio not in prio_sched_lists:
                prio_sched_lists[sched_event.prio] = []

            prio_sched_lists[sched_event.prio].append(sched_event)

        for prio in prio_sched_lists:
            sched_list = prio_sched_lists[prio]

            if not sched_list:
                continue

            stdev = self._compute_sched_latency_stdev(sched_list)
            latencies = [sched.latency for sched in sched_list]
            count = len(latencies)
            min_latency = min(latencies)
            max_latency = max(latencies)
            total_latency = sum(latencies)

            prio_stats[prio] = _SchedStats(
                count=count,
                min=min_latency,
                max=max_latency,
                stdev=stdev,
                total=total_latency,
            )

        return prio_sched_lists, prio_stats

    def _get_log_result_table(self, period_data, begin_ns, end_ns):
        result_table = self._mi_create_result_table(self._MI_TABLE_CLASS_LOG,
                                                    begin_ns, end_ns)

        for sched_event in period_data.sched_list:
            wakee_proc = mi.Process(sched_event.wakee_proc.comm,
                                    sched_event.wakee_proc.pid,
                                    sched_event.wakee_proc.tid)

            if sched_event.waker_proc:
                waker_proc = mi.Process(sched_event.waker_proc.comm,
                                        sched_event.waker_proc.pid,
                                        sched_event.waker_proc.tid)
            else:
                waker_proc = mi.Empty()

            result_table.append_row(
                wakeup_ts=mi.Timestamp(sched_event.wakeup_ts),
                switch_ts=mi.Timestamp(sched_event.switch_ts),
                latency=mi.Duration(sched_event.latency),
                prio=mi.Number(sched_event.prio),
                target_cpu=mi.Cpu(sched_event.target_cpu),
                wakee_proc=wakee_proc,
                waker_proc=waker_proc,
            )

        return result_table

    def _get_top_result_table(self, period_data, begin_ns, end_ns):
        result_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_TOP, begin_ns, end_ns)

        top_events = sorted(period_data.sched_list,
                            key=operator.attrgetter('latency'),
                            reverse=True)
        top_events = top_events[:self._args.limit]

        for sched_event in top_events:
            wakee_proc = mi.Process(sched_event.wakee_proc.comm,
                                    sched_event.wakee_proc.pid,
                                    sched_event.wakee_proc.tid)

            if sched_event.waker_proc:
                waker_proc = mi.Process(sched_event.waker_proc.comm,
                                        sched_event.waker_proc.pid,
                                        sched_event.waker_proc.tid)
            else:
                waker_proc = mi.Empty()

            result_table.append_row(
                wakeup_ts=mi.Timestamp(sched_event.wakeup_ts),
                switch_ts=mi.Timestamp(sched_event.switch_ts),
                latency=mi.Duration(sched_event.latency),
                prio=mi.Number(sched_event.prio),
                target_cpu=mi.Cpu(sched_event.target_cpu),
                wakee_proc=wakee_proc,
                waker_proc=waker_proc,
            )

        return result_table

    def _get_total_stats_result_table(self, period_data, begin_ns, end_ns):
        stats_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_TOTAL_STATS,
                                         begin_ns, end_ns)

        stdev = self._compute_sched_latency_stdev(period_data.sched_list)
        if math.isnan(stdev):
            stdev = mi.Unknown()
        else:
            stdev = mi.Duration(stdev)

        count = self._analysis.count(period_data)
        if count == 0:
            avg = mi.Duration(0)
        else:
            avg = mi.Duration(period_data.total_latency / count)

        if period_data.min_latency is None:
            min = mi.Duration(0)
        else:
            min = mi.Duration(period_data.min_latency)

        if period_data.max_latency is None:
            max = mi.Duration(0)
        else:
            max = mi.Duration(period_data.max_latency)

        stats_table.append_row(
            count=mi.Number(self._analysis.count(period_data)),
            min_latency=min,
            avg_latency=avg,
            max_latency=max,
            stdev_latency=stdev,
        )

        return stats_table

    def _get_per_tid_stats_result_table(self, period_data, begin_ns, end_ns):
        stats_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_PER_TID_STATS,
                                         begin_ns, end_ns)

        tid_stats_list = sorted(list(period_data.tids.values()),
                                key=lambda proc: proc.comm.lower())

        for tid_stats in tid_stats_list:
            if not tid_stats.sched_list:
                continue

            stdev = self._compute_sched_latency_stdev(tid_stats.sched_list)
            if math.isnan(stdev):
                stdev = mi.Unknown()
            else:
                stdev = mi.Duration(stdev)

            prio_list = format_utils.format_prio_list(tid_stats.prio_list)

            stats_table.append_row(
                process=mi.Process(tid=tid_stats.tid, name=tid_stats.comm),
                count=mi.Number(tid_stats.count),
                min_latency=mi.Duration(tid_stats.min_latency),
                avg_latency=mi.Duration(tid_stats.total_latency /
                                        tid_stats.count),
                max_latency=mi.Duration(tid_stats.max_latency),
                stdev_latency=stdev,
                prio_list=mi.String(prio_list),
            )

        return stats_table

    def _get_per_prio_stats_result_table(self, period_data, begin_ns, end_ns):
        stats_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_PER_PRIO_STATS,
                                         begin_ns, end_ns)

        _, prio_stats = self._get_prio_sched_lists_stats(period_data)

        for prio in sorted(prio_stats):
            stats = prio_stats[prio]
            stdev = stats.stdev

            if math.isnan(stdev):
                stdev = mi.Unknown()
            else:
                stdev = mi.Duration(stdev)

            count = stats.count
            min_latency = stats.min
            max_latency = stats.max
            total_latency = stats.total

            stats_table.append_row(
                prio=mi.Number(prio),
                count=mi.Number(count),
                min_latency=mi.Duration(min_latency),
                avg_latency=mi.Duration(total_latency / count),
                max_latency=mi.Duration(max_latency),
                stdev_latency=stdev,
            )

        return stats_table

    def _get_per_tid_freq_series_table(self, freq_tables):
        if not freq_tables:
            return

        column_infos = [
            ('duration_lower', 'Duration (lower bound)', mi.Duration),
            ('duration_upper', 'Duration (upper bound)', mi.Duration),
        ]

        for index, freq_table in enumerate(freq_tables):
            column_infos.append((
                'tid{}'.format(index),
                freq_table.subtitle,
                mi.Number,
                'schedulings'
            ))

        title = 'Scheduling latencies frequency distributions'
        table_class = mi.TableClass(None, title, column_infos)
        begin = freq_tables[0].timerange.begin.value
        end = freq_tables[0].timerange.end.value
        result_table = mi.ResultTable(table_class, begin, end)

        for row_index, freq0_row in enumerate(freq_tables[0].rows):
            row_tuple = [
                freq0_row.duration_lower,
                freq0_row.duration_upper,
            ]

            for freq_table in freq_tables:
                freq_row = freq_table.rows[row_index]
                row_tuple.append(freq_row.count)

            result_table.append_row_tuple(tuple(row_tuple))

        return result_table

    def _get_per_prio_freq_series_table(self, freq_tables):
        if not freq_tables:
            return

        column_infos = [
            ('duration_lower', 'Duration (lower bound)', mi.Duration),
            ('duration_upper', 'Duration (upper bound)', mi.Duration),
        ]

        for index, freq_table in enumerate(freq_tables):
            column_infos.append((
                'prio{}'.format(index),
                freq_table.subtitle,
                mi.Number,
                'schedulings'
            ))

        title = 'Scheduling latencies frequency distributions'
        table_class = mi.TableClass(None, title, column_infos)
        begin = freq_tables[0].timerange.begin.value
        end = freq_tables[0].timerange.end.value
        result_table = mi.ResultTable(table_class, begin, end)

        for row_index, freq0_row in enumerate(freq_tables[0].rows):
            row_tuple = [
                freq0_row.duration_lower,
                freq0_row.duration_upper,
            ]

            for freq_table in freq_tables:
                freq_row = freq_table.rows[row_index]
                row_tuple.append(freq_row.count)

            result_table.append_row_tuple(tuple(row_tuple))

        return result_table

    def _fill_freq_result_table(self, sched_list, stats, min_duration,
                                max_duration, step, freq_table):
        # The number of bins for the histogram
        resolution = self._args.freq_resolution

        if not self._args.freq_uniform:
            if self._args.min is not None:
                min_duration = self._args.min
            else:
                min_duration = stats.min

            if self._args.max is not None:
                max_duration = self._args.max
            else:
                max_duration = stats.max

            # ns to µs
            if min_duration is None:
                min_duration = 0
            else:
                min_duration /= 1000

            if max_duration is None:
                max_duration = 0
            else:
                max_duration /= 1000

            step = (max_duration - min_duration) / resolution

        if step == 0:
            return

        buckets = []
        counts = []

        for i in range(resolution):
            buckets.append(i * step)
            counts.append(0)

        for sched_event in sched_list:
            duration = sched_event.latency / 1000
            index = int((duration - min_duration) / step)

            if index >= resolution:
                # special case for max value: put in last bucket (includes
                # its upper bound)
                if duration == max_duration:
                    counts[index - 1] += 1

                continue

            counts[index] += 1

        for index, count in enumerate(counts):
            lower_bound = index * step + min_duration
            upper_bound = (index + 1) * step + min_duration
            freq_table.append_row(
                duration_lower=mi.Duration.from_us(lower_bound),
                duration_upper=mi.Duration.from_us(upper_bound),
                count=mi.Number(count),
            )

    def _get_total_freq_result_tables(self, period_data, begin_ns, end_ns):
        freq_tables = []
        sched_lists, sched_stats = self._get_total_sched_lists_stats(
            period_data)
        min_duration = None
        max_duration = None
        step = None

        if self._args.freq_uniform:
            latencies = []

            for sched_list in sched_lists:
                latencies += [sched.latency for sched in sched_list]

            min_duration, max_duration, step = \
                self._find_uniform_freq_values(latencies)

        for sched_list in sched_lists:
            freq_table = \
                self._mi_create_result_table(self._MI_TABLE_CLASS_FREQ,
                                             begin_ns, end_ns)
            self._fill_freq_result_table(sched_list, sched_stats, min_duration,
                                         max_duration, step, freq_table)
            freq_tables.append(freq_table)

        return freq_tables

    def _get_per_tid_freq_result_tables(self, period_data, begin_ns, end_ns):
        freq_tables = []
        tid_sched_lists, tid_stats = self._get_tid_sched_lists_stats(
            period_data)
        min_duration = None
        max_duration = None
        step = None

        if self._args.freq_uniform:
            latencies = []

            for sched_list in tid_sched_lists.values():
                latencies += [sched.latency for sched in sched_list]

            min_duration, max_duration, step = \
                self._find_uniform_freq_values(latencies)

        for tid in sorted(tid_sched_lists):
            sched_list = tid_sched_lists[tid]
            stats = tid_stats[tid]
            subtitle = 'TID: {}'.format(tid)
            freq_table = \
                self._mi_create_result_table(self._MI_TABLE_CLASS_FREQ,
                                             begin_ns, end_ns, subtitle)
            self._fill_freq_result_table(sched_list, stats, min_duration,
                                         max_duration, step, freq_table)
            freq_tables.append(freq_table)

        return freq_tables

    def _get_per_prio_freq_result_tables(self, period_data, begin_ns, end_ns):
        freq_tables = []
        prio_sched_lists, prio_stats = self._get_prio_sched_lists_stats(
            period_data)
        min_duration = None
        max_duration = None
        step = None

        if self._args.freq_uniform:
            latencies = []

            for sched_list in prio_sched_lists.values():
                latencies += [sched.latency for sched in sched_list]

            min_duration, max_duration, step = \
                self._find_uniform_freq_values(latencies)

        for prio in sorted(prio_sched_lists):
            sched_list = prio_sched_lists[prio]
            stats = prio_stats[prio]
            subtitle = 'Priority: {}'.format(prio)
            freq_table = \
                self._mi_create_result_table(self._MI_TABLE_CLASS_FREQ,
                                             begin_ns, end_ns, subtitle)
            self._fill_freq_result_table(sched_list, stats, min_duration,
                                         max_duration, step, freq_table)
            freq_tables.append(freq_table)

        return freq_tables

    def _compute_sched_latency_stdev(self, sched_events):
        sched_latencies = []
        for sched_event in sched_events:
            sched_latencies.append(sched_event.latency)
        if len(sched_latencies) < 2:
            return float('nan')
        return statistics.stdev(sched_latencies)

    def _print_sched_events(self, result_table):
        fmt = '[{:<18}, {:<18}] {:>15} {:>10}  {:>3}   {:<25}  {:<25}'
        title_fmt = '{:<20} {:<19} {:>15} {:>10}  {:>3}   {:<25}  {:<25}'
        print()
        print(result_table.title)
        print(title_fmt.format('Wakeup', 'Switch', 'Latency (us)', 'Priority',
                               'CPU', 'Wakee', 'Waker'))
        for row in result_table.rows:
            wakeup_ts = row.wakeup_ts.value
            switch_ts = row.switch_ts.value
            latency = row.latency.value
            prio = row.prio.value
            target_cpu = row.target_cpu.id
            wakee_proc = row.wakee_proc
            waker_proc = row.waker_proc

            wakee_str = '%s (%d)' % (wakee_proc.name, wakee_proc.tid)
            if isinstance(waker_proc, mi.Empty):
                waker_str = 'Unknown (N/A)'
            else:
                waker_str = '%s (%d)' % (waker_proc.name, waker_proc.tid)

            print(fmt.format(self._format_timestamp(wakeup_ts),
                             self._format_timestamp(switch_ts),
                             '%0.03f' % (latency / 1000), prio,
                             target_cpu, wakee_str, waker_str))

    def _print_total_stats(self, stats_table):
        row_format = '{:<12} {:<12} {:<12} {:<12} {:<12}'
        header = row_format.format(
            'Count', 'Min', 'Avg', 'Max', 'Stdev'
        )

        if stats_table.rows:
            print()
            print(stats_table.title + ' (us)')
            print(header)
            for row in stats_table.rows:
                if type(row.stdev_latency) is mi.Unknown:
                    stdev_str = '?'
                else:
                    stdev_str = '%0.03f' % row.stdev_latency.to_us()

                row_str = row_format.format(
                    '%d' % row.count.value,
                    '%0.03f' % row.min_latency.to_us(),
                    '%0.03f' % row.avg_latency.to_us(),
                    '%0.03f' % row.max_latency.to_us(),
                    '%s' % stdev_str,
                )

                print(row_str)

    def _print_per_tid_stats(self, stats_table):
        row_format = '{:<25} {:>8}  {:>12}  {:>12}  {:>12}  {:>12}   {}'
        header = row_format.format(
            'Process', 'Count', 'Min', 'Avg', 'Max', 'Stdev', 'Priorities'
        )

        if stats_table.rows:
            print()
            print(stats_table.title + ' (us)')
            print(header)
            for row in stats_table.rows:
                if type(row.stdev_latency) is mi.Unknown:
                    stdev_str = '?'
                else:
                    stdev_str = '%0.03f' % row.stdev_latency.to_us()

                proc = row.process
                proc_str = '%s (%d)' % (proc.name, proc.tid)

                row_str = row_format.format(
                    '%s' % proc_str,
                    '%d' % row.count.value,
                    '%0.03f' % row.min_latency.to_us(),
                    '%0.03f' % row.avg_latency.to_us(),
                    '%0.03f' % row.max_latency.to_us(),
                    '%s' % stdev_str,
                    '%s' % row.prio_list.value,
                )

                print(row_str)

    def _print_per_prio_stats(self, stats_table):
        row_format = '{:>4} {:>8}  {:>12}  {:>12}  {:>12}  {:>12}'
        header = row_format.format(
            'Prio', 'Count', 'Min', 'Avg', 'Max', 'Stdev'
        )

        if stats_table.rows:
            print()
            print(stats_table.title + ' (us)')
            print(header)
            for row in stats_table.rows:
                if type(row.stdev_latency) is mi.Unknown:
                    stdev_str = '?'
                else:
                    stdev_str = '%0.03f' % row.stdev_latency.to_us()

                row_str = row_format.format(
                    '%d' % row.prio.value,
                    '%d' % row.count.value,
                    '%0.03f' % row.min_latency.to_us(),
                    '%0.03f' % row.avg_latency.to_us(),
                    '%0.03f' % row.max_latency.to_us(),
                    '%s' % stdev_str,
                )

                print(row_str)

    def _print_frequency_distribution(self, freq_table):
        title_fmt = 'Scheduling latency frequency distribution - {}'

        graph = termgraph.FreqGraph(
            data=freq_table.rows,
            get_value=lambda row: row.count.value,
            get_lower_bound=lambda row: row.duration_lower.to_us(),
            title=title_fmt.format(freq_table.subtitle),
            unit='µs'
        )

        graph.print_graph()

    def _print_freq(self, freq_tables):
        for freq_table in freq_tables:
            self._print_frequency_distribution(freq_table)

    def _validate_transform_args(self):
        args = self._args

        # If neither --total nor --per-prio are specified, default
        # to --per-tid
        if not (args.total or args.per_prio):
            args.per_tid = True

    def _add_arguments(self, ap):
        Command._add_min_max_args(ap)
        Command._add_proc_filter_args(ap)
        Command._add_freq_args(
            ap, help='Output the frequency distribution of sched switch '
            'latencies')
        Command._add_top_args(ap, help='Output the top sched switch latencies')
        Command._add_log_args(
            ap, help='Output the sched switches in chronological order')
        Command._add_stats_args(ap, help='Output sched switch statistics')
        ap.add_argument('--total', action='store_true',
                        help='Group all results (applies to stats and freq)')
        ap.add_argument('--per-tid', action='store_true',
                        help='Group results per-TID (applies to stats and '
                        'freq) (default)')
        ap.add_argument('--per-prio', action='store_true',
                        help='Group results per-prio (applies to stats and '
                        'freq)')


def _run(mi_mode):
    schedcmd = SchedAnalysisCommand(mi_mode=mi_mode)
    schedcmd.run()


def _runstats(mi_mode):
    sys.argv.insert(1, '--stats')
    _run(mi_mode)


def _runlog(mi_mode):
    sys.argv.insert(1, '--log')
    _run(mi_mode)


def _runtop(mi_mode):
    sys.argv.insert(1, '--top')
    _run(mi_mode)


def _runfreq(mi_mode):
    sys.argv.insert(1, '--freq')
    _run(mi_mode)


def runstats():
    _runstats(mi_mode=False)


def runlog():
    _runlog(mi_mode=False)


def runtop():
    _runtop(mi_mode=False)


def runfreq():
    _runfreq(mi_mode=False)


def runstats_mi():
    _runstats(mi_mode=True)


def runlog_mi():
    _runlog(mi_mode=True)


def runtop_mi():
    _runtop(mi_mode=True)


def runfreq_mi():
    _runfreq(mi_mode=True)



# The MIT License (MIT)
#
# Copyright (C) 2016 - Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import babeltrace as bt
import collections


_CTF_SCOPES = (
    bt.CTFScope.EVENT_FIELDS,
    bt.CTFScope.EVENT_CONTEXT,
    bt.CTFScope.STREAM_EVENT_CONTEXT,
    bt.CTFScope.STREAM_EVENT_HEADER,
    bt.CTFScope.STREAM_PACKET_CONTEXT,
    bt.CTFScope.TRACE_PACKET_HEADER,
)


# This class has an interface which is compatible with the
# babeltrace.reader.Event class. This is the result of a deep copy
# performed by LTTng analyses.
class Event(collections.Mapping):
    def __init__(self, bt_ev):
        self._copy_bt_event(bt_ev)

    def _copy_bt_event(self, bt_ev):
        self._name = bt_ev.name
        self._cycles = bt_ev.cycles
        self._timestamp = bt_ev.timestamp
        self._fields = {}

        for scope in _CTF_SCOPES:
            self._fields[scope] = {}

            for field_name in bt_ev.field_list_with_scope(scope):
                field_value = bt_ev.field_with_scope(field_name, scope)
                self._fields[scope][field_name] = field_value

    @property
    def name(self):
        return self._name

    @property
    def cycles(self):
        return self._cycles

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def handle(self):
        raise NotImplementedError()

    @property
    def trace_collection(self):
        raise NotImplementedError()

    def _get_first_field(self, field_name):
        for scope_fields in self._fields.values():
            if field_name in scope_fields:
                return scope_fields[field_name]

    def field_with_scope(self, field_name, scope):
        if scope not in self._fields:
            raise ValueError('Invalid scope provided')

        if field_name in self._fields[scope]:
            return self._fields[scope][field_name]

    def field_list_with_scope(self, scope):
        if scope not in self._fields:
            raise ValueError('Invalid scope provided')

        return list(self._fields[scope].keys())

    def __getitem__(self, field_name):
        field = self._get_first_field(field_name)

        if field is None:
            raise KeyError(field_name)

        return field

    def __iter__(self):
        for key in self.keys():
            yield key

    def __len__(self):
        count = 0

        for scope_fields in self._fields.values():
            count += len(scope_fields)

        return count

    def __contains__(self, field_name):
        return self._get_first_field(field_name) is not None

    def keys(self):
        keys = []

        for scope_fields in self._fields.values():
            keys += list(scope_fields.keys())

        return keys

    def get(self, field_name, default=None):
        field = self._get_first_field(field_name)

        if field is None:
            return default

        return field

    def items(self):
        raise NotImplementedError()



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2016 - Philippe Proulx <pproulx@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import argparse
import json
import os
import re
import sys
import subprocess
import traceback
from babeltrace import TraceCollection
from . import mi, progressbar, period_parsing
from .. import __version__
from ..core import analysis, period as core_period
from ..common import (
    format_utils, parse_utils, trace_utils, version_utils
)
from ..linuxautomaton import automaton


class Command:
    _MI_BASE_TAGS = ['linux-kernel', 'lttng-analyses']
    _MI_AUTHORS = [
        'Julien Desfossez',
        'Antoine Busque',
        'Philippe Proulx',
    ]
    _MI_URL = 'https://github.com/lttng/lttng-analyses'
    _VERSION = version_utils.Version.new_from_string(__version__)
    _BT_INTERSECT_VERSION = version_utils.Version(1, 4, 0)
    _DEBUG_ENV_VAR = 'LTTNG_ANALYSES_DEBUG'

    def __init__(self, mi_mode=False):
        self._analysis = None
        self._analysis_conf = None
        self._args = None
        self._babeltrace_version = None
        self._handles = None
        self._traces = None
        self._period_ticks = 0
        self._mi_mode = mi_mode
        self._debug_mode = os.environ.get(self._DEBUG_ENV_VAR)
        self._run_step('create automaton', self._create_automaton)
        self._run_step('setup MI', self._mi_setup)

    @property
    def mi_mode(self):
        return self._mi_mode

    def _run_step(self, action_title, fn):
        try:
            fn()
        except KeyboardInterrupt:
            self._print('Cancelled by user')
            sys.exit(0)
        except Exception as e:
            self._gen_error('Cannot {}: {}'.format(action_title, e))

    def run(self):
        self._run_step('parse arguments', self._parse_args)
        self._run_step('open trace', self._open_trace)
        self._run_step('create analysis', self._create_analysis)

        if not self._mi_mode or not self._args.test_compatibility:
            self._run_step('run analysis', self._run_analysis)

        self._run_step('close trace', self._close_trace)

    def _mi_error(self, msg, code=None):
        print(json.dumps(mi.get_error(msg, code)))

    def _non_mi_msg(self, msg, color):
        if self._args.color:
            try:
                import termcolor

                msg = termcolor.colored(msg, color, attrs=['bold'])
            except ImportError:
                pass

        print(msg, file=sys.stderr)

    def _non_mi_error(self, msg):
        self._non_mi_msg(msg, 'red')

    def _non_mi_warn(self, msg):
        self._non_mi_msg(msg, 'yellow')

    def _error(self, msg, exit_code=1):
        if self._debug_mode:
            traceback.print_exc()

        if self._mi_mode:
            self._mi_error(msg)
        else:
            self._non_mi_error(msg)

        if exit_code is not None:
            sys.exit(exit_code)

    def _warn(self, msg):
        if not self._mi_mode:
            self._non_mi_warn(msg)

    def _gen_error(self, msg, exit_code=1):
        self._error('Error: {}'.format(msg), exit_code)

    def _cmdline_error(self, msg, exit_code=1):
        self._error('Command line error: {}'.format(msg), exit_code)

    def _print(self, msg):
        if not self._mi_mode:
            print(msg)

    def _mi_create_result_table(self, table_class_name, begin, end,
                                subtitle=None):
        return mi.ResultTable(self._mi_table_classes[table_class_name],
                              begin, end, subtitle)

    def _mi_setup(self):
        self._mi_table_classes = {}

        for tc_tuple in self._MI_TABLE_CLASSES:
            table_class = mi.TableClass(tc_tuple[0], tc_tuple[1], tc_tuple[2])
            self._mi_table_classes[table_class.name] = table_class

        self._mi_clear_result_tables()

    def _mi_print_metadata(self):
        tags = self._MI_BASE_TAGS + self._MI_TAGS
        infos = mi.get_metadata(version=self._VERSION, title=self._MI_TITLE,
                                description=self._MI_DESCRIPTION,
                                authors=self._MI_AUTHORS, url=self._MI_URL,
                                tags=tags,
                                table_classes=self._mi_table_classes.values())
        print(json.dumps(infos))

    def _mi_append_result_table(self, result_table):
        if not result_table or not result_table.rows:
            return

        tc_name = result_table.table_class.name
        self._mi_get_result_tables(tc_name).append(result_table)

    def _mi_append_result_tables(self, result_tables):
        if not result_tables:
            return

        for result_table in result_tables:
            self._mi_append_result_table(result_table)

    def _mi_clear_result_tables(self):
        self._result_tables = {}

    def _mi_get_result_tables(self, table_class_name):
        if table_class_name not in self._result_tables:
            self._result_tables[table_class_name] = []

        return self._result_tables[table_class_name]

    def _mi_print(self):
        results = []

        for result_tables in self._result_tables.values():
            for result_table in result_tables:
                results.append(result_table.to_native_object())

        obj = {
            'results': results,
        }

        print(json.dumps(obj))

    def _create_summary_result_tables(self):
        pass

    def _open_trace(self):
        self._babeltrace_version = trace_utils.read_babeltrace_version()
        if self._babeltrace_version >= self._BT_INTERSECT_VERSION:
            traces = TraceCollection(intersect_mode=self._args.intersect_mode)
        else:
            if self._args.intersect_mode:
                self._print('Warning: intersect mode not available - '
                            'disabling')
                self._print('         Use babeltrace {} or later to '
                            'enable'.format(
                                trace_utils.BT_INTERSECT_VERSION))
                self._args.intersect_mode = False
            traces = TraceCollection()
        handles = traces.add_traces_recursive(self._args.path, 'ctf')
        if handles == {}:
            self._gen_error('Failed to open ' + self._args.path, -1)
        self._handles = handles
        self._traces = traces
        self._ts_begin = traces.timestamp_begin
        self._ts_end = traces.timestamp_end
        self._process_date_args()
        self._read_tracer_version()
        if not self._args.skip_validation:
            self._check_lost_events()
        if not self._check_period_args():
            self._gen_error('Invalid period parameters')

    def _close_trace(self):
        for handle in self._handles.values():
            self._traces.remove_trace(handle)

    def _read_tracer_version(self):
        # TODO: associate the version of the tracer with each trace, not
        # globally. Waiting for bug #1085 to be fixed in Babeltrace.
        kernel_path = None
        # remove the trailing /
        while self._args.path.endswith('/'):
            self._args.path = self._args.path[:-1]
        for root, _, _ in os.walk(self._args.path):
            if root.endswith('kernel'):
                kernel_path = root
                break

        # If we don't have a kernel folder, we don't need to check the version
        # of the tracer for now.
        if kernel_path is None:
            return

        try:
            ret, metadata = subprocess.getstatusoutput(
                'babeltrace -o ctf-metadata "%s"' % kernel_path)
        except subprocess.CalledProcessError:
            self._gen_error('Cannot run babeltrace on the trace, cannot read'
                            ' tracer version')

        # fallback to reading the text metadata if babeltrace failed to
        # output the CTF metadata
        if ret != 0:
            try:
                metadata = subprocess.getoutput(
                    'cat "%s"' % os.path.join(kernel_path, 'metadata'))
            except subprocess.CalledProcessError:
                self._gen_error('Cannot read the metadata of the trace, cannot'
                                'extract tracer version')

        major_match = re.search(r'tracer_major = "*(\d+)"*', metadata)
        minor_match = re.search(r'tracer_minor = "*(\d+)"*', metadata)
        patch_match = re.search(r'tracer_patchlevel = "*(\d+)"*', metadata)

        if not major_match or not minor_match or not patch_match:
            self._gen_error('Malformed metadata, cannot read tracer version')

        self.state.tracer_version = version_utils.Version(
            int(major_match.group(1)),
            int(minor_match.group(1)),
            int(patch_match.group(1)),
        )

    def _read_babeltrace_version(self):
        try:
            output = subprocess.check_output('babeltrace')
        except subprocess.CalledProcessError:
            self._gen_error('Could not run babeltrace to verify version')

        output = output.decode(sys.stdout.encoding)
        first_line = output.splitlines()[0]
        version_string = first_line.split()[-1]

        self._babeltrace_version = \
            version_utils.Version.new_from_string(version_string)

    def _check_lost_events(self):
        msg = 'Checking the trace for lost events...'
        self._print(msg)

        if self._mi_mode and self._args.output_progress:
            mi.print_progress(0, msg)

        try:
            subprocess.check_output('babeltrace "%s"' % self._args.path,
                                    shell=True)
        except subprocess.CalledProcessError:
            self._gen_error('Cannot run babeltrace on the trace, cannot verify'
                            ' if events were lost during the trace recording')

    def _pre_analysis(self):
        pass

    def _mi_post_analysis(self):
        if not self._mi_mode:
            return

        if self._period_ticks > 1:
            self._create_summary_result_tables()

        self._mi_print()

    def _post_analysis(self):
        self._mi_post_analysis()

    def _pb_setup(self):
        if self._args.no_progress:
            return

        ts_end = self._ts_end

        if self._analysis_conf.end_ts is not None:
            ts_end = self._analysis_conf.end_ts

        if self._mi_mode:
            cls = progressbar.MiProgress
        else:
            cls = progressbar.FancyProgressBar

        self._progress = cls(self._ts_begin, ts_end, self._args.path,
                             self._args.progress_use_size)

    def _pb_update(self, event):
        if self._args.no_progress:
            return

        self._progress.update(event)

    def _pb_finish(self):
        if self._args.no_progress:
            return

        self._progress.finalize()

    def _run_analysis(self):
        self._pre_analysis()
        self._pb_setup()

        if self._args.intersect_mode:
            if not self._traces.has_intersection:
                self._gen_error('Trace has no intersection. '
                                'Use --no-intersection to override')

        first_event = True
        for event in self._traces.events:
            if first_event is True:
                self._analysis.begin_analysis(event)
                first_event = False
            self._pb_update(event)
            self._analysis.process_event(event)
            if self._analysis.ended:
                break
            self._automaton.process_event(event)

        self._pb_finish()
        self._analysis.end_analysis()
        self._post_analysis()

    def _print_date(self, begin_ns, end_ns):
        time_range_str = format_utils.format_time_range(
            begin_ns, end_ns, print_date=True, gmt=self._args.gmt
        )
        date = 'Timerange: {}'.format(time_range_str)

        self._print(date)

    def _format_timestamp(self, timestamp):
        return format_utils.format_timestamp(
            timestamp, print_date=self._args.multi_day, gmt=self._args.gmt
        )

    def _uniform_freq_min(self, category='default'):
        return self._analysis_conf.uniform_min[category]

    def _uniform_freq_max(self, category='default'):
        return self._analysis_conf.uniform_max[category]

    def _uniform_freq_step(self, category='default'):
        return self._analysis_conf.uniform_step[category]

    def _find_uniform_freq_values(self, durations, ratio=1000,
                                  category='default'):
        if category not in self._analysis_conf.uniform_step.keys():
            self._analysis_conf.uniform_min[category] = None
            self._analysis_conf.uniform_max[category] = None
            self._analysis_conf.uniform_step[category] = None

        if self._args.min is not None:
            self._analysis_conf.uniform_min[category] = self._args.min
        else:
            if len(durations) == 0:
                self._analysis_conf.uniform_min[category] = 0
            else:
                if self._analysis_conf.uniform_min[category] is None or \
                        min(durations) / ratio < \
                        self._analysis_conf.uniform_min[category]:
                    self._analysis_conf.uniform_min[category] = \
                        min(durations) / ratio
        if self._args.max is not None:
            self._analysis_conf.uniform_max[category] = self._args.max
        else:
            if len(durations) == 0:
                self._analysis_conf.uniform_max[category] = 0
            else:
                if self._analysis_conf.uniform_max[category] is None or \
                        max(durations) / ratio > \
                        self._analysis_conf.uniform_max[category]:
                    self._analysis_conf.uniform_max[category] = \
                        max(durations) / ratio

        # ns to µs
        self._analysis_conf.uniform_step[category] = (
            (self._analysis_conf.uniform_max[category] -
                self._analysis_conf.uniform_min[category]) /
            self._args.freq_resolution)

        return self._analysis_conf.uniform_min[category], \
            self._analysis_conf.uniform_max[category], \
            self._analysis_conf.uniform_step[category]

    def _check_period_args(self):
        # FIXME
        return True

        if len(self._analysis_conf.period_defs) > 0:
            name = self._analysis_conf.period_begin_ev_name
            if not trace_utils.check_event_exists(self._handles, name):
                self._gen_error("Event %s not found in the trace" % name)
                return False
        if self._analysis_conf.period_end_ev_name is not None:
            name = self._analysis_conf.period_end_ev_name
            if not trace_utils.check_event_exists(self._handles, name):
                self._gen_error("Event %s not found in the trace" % name)
                return False

        ev_name = self._analysis_conf.period_begin_ev_name
        for field in self._analysis_conf.period_begin_key_fields:
            if not ev_name:
                break
            if not trace_utils.check_field_exists(self._handles,
                                                  ev_name, field):
                self._gen_error("Field %s not found in event %s" %
                                (field, ev_name))
                return False
        ev_name = self._analysis_conf.period_end_ev_name
        for field in self._analysis_conf.period_end_key_fields:
            if not ev_name:
                break
            if not trace_utils.check_field_exists(self._handles,
                                                  ev_name, field):
                self._gen_error("Field %s not found in event %s" %
                                (field, ev_name))
                return False
        return True

    def _validate_transform_period_args(self, analysis_conf):
        args = self._args

        # validate period arguments
        if (args.period_begin is not None or args.period_end is not None or
           args.period_key_value is not None or
           args.period_begin_key is not None or
           args.period_end_key is not None) and (args.period or
                                                 args.period_captures):
            self._cmdline_error('Do not use another period option when using '
                                'one or more --period or --period-captures '
                                'options')

        registry = self._analysis_conf.period_def_registry
        name_to_begin_captures_exprs = {}
        name_to_end_captures_exprs = {}

        # parse period definition expressions
        if args.period:
            # period captures first
            if args.period_captures:
                for arg in args.period_captures:
                    try:
                        res = period_parsing.parse_period_captures_arg(arg)
                    except period_parsing.MalformedExpression as e:
                        self._cmdline_error('Malformed period captures '
                                            'expression: {}'.format(e))
                    except Exception as e:
                        self._cmdline_error('Cannot parse period captures '
                                            'expression: {}'.format(e))

                    if res.name in name_to_begin_captures_exprs:
                        fmt = 'Duplicate period name "{}" in ' \
                              '--period-captures argument'
                        self._cmdline_error(fmt.format(res.name))

                    name_to_begin_captures_exprs[res.name] = \
                        res.begin_captures_exprs
                    name_to_end_captures_exprs[res.name] = \
                        res.end_captures_exprs

            for period_arg in args.period:
                try:
                    res = period_parsing.parse_period_def_arg(period_arg)
                except period_parsing.MalformedExpression as e:
                    self._cmdline_error('Malformed period definition '
                                        'expression: {}'.format(e))
                except Exception as e:
                    self._cmdline_error('Cannot parse period definition '
                                        'expression: {}'.format(e))

                begin_captures_exprs = {}
                end_captures_exprs = {}

                if res.period_name is not None:
                    begin_captures_exprs = name_to_begin_captures_exprs.get(
                        res.period_name, {})
                    end_captures_exprs = name_to_end_captures_exprs.get(
                        res.period_name, {})

                try:
                    registry.add_period_def(res.parent_name, res.period_name,
                                            res.begin_expr, res.end_expr,
                                            begin_captures_exprs,
                                            end_captures_exprs)
                except core_period.IllegalExpression as e:
                    self._cmdline_error('Illegal period definition '
                                        'expression: {}'.format(e))
                except core_period.InvalidPeriodDefinition as e:
                    self._cmdline_error('Cannot add period: {}'.format(e),
                                        None)
                    self._error('Period argument: {}'.format(period_arg))
        elif (args.period_begin is not None or args.period_end is not None):
            self._warn('''Warning: The following period options are deprecated:

  --period-begin
  --period-end
  --period-begin-key
  --period-end-key
  --period-key-value

Please consider using the --period option.''')

            # create new-style expression from old-style arguments
            if args.period_begin is None:
                # ignore incomplete period definition
                return

            if not args.period_begin_key:
                args.period_begin_key = 'cpu_id'

            if not args.period_end:
                args.period_end = args.period_begin

            if not args.period_end_key:
                args.period_end_key = args.period_begin_key

            begin_exprs = []
            end_exprs = []

            # conditions for matching the event name
            begin_event_name_expr = core_period.EventScope(
                core_period.EventName())
            begin_event_name_str_expr = core_period.String(args.period_begin)
            end_event_name_expr = core_period.EventScope(
                core_period.EventName())
            end_event_name_str_expr = core_period.String(args.period_end)
            begin_event_name_eq_expr = core_period.Eq(
                begin_event_name_expr, begin_event_name_str_expr)
            begin_exprs.append(begin_event_name_eq_expr)
            end_event_name_eq_expr = core_period.Eq(end_event_name_expr,
                                                    end_event_name_str_expr)
            end_exprs.append(end_event_name_eq_expr)
            begin_field_names = args.period_begin_key.split(',')
            end_field_names = args.period_end_key.split(',')

            # conditions for begin field values
            if args.period_key_value:
                parts = args.period_key_value.split(',')
                value_exprs = []

                for part in parts:
                    try:
                        value_exprs.append(core_period.Number(float(part)))
                    except:
                        value_exprs.append(core_period.String(part))

                for field_name, value_expr in zip(begin_field_names,
                                                  value_exprs):
                    event_field_name = core_period.EventFieldName(field_name)
                    dyn_scope = core_period.DynamicScope(
                        core_period.DynScope.AUTO, event_field_name)
                    evt_scope = core_period.EventScope(dyn_scope)
                    begin_scope = core_period.BeginScope(evt_scope)
                    eq_expr = core_period.Eq(begin_scope, value_expr)
                    begin_exprs.append(eq_expr)

            # conditions for equal end and begin fields
            for begin_field_name, end_field_name in zip(begin_field_names,
                                                        end_field_names):
                # begin scope
                event_field_name = core_period.EventFieldName(begin_field_name)
                dyn_scope = core_period.DynamicScope(core_period.DynScope.AUTO,
                                                     event_field_name)
                evt_scope = core_period.EventScope(dyn_scope)
                begin_scope = core_period.BeginScope(evt_scope)

                # end event scope
                event_field_name = core_period.EventFieldName(end_field_name)
                dyn_scope = core_period.DynamicScope(core_period.DynScope.AUTO,
                                                     event_field_name)
                end_evt_scope = core_period.EventScope(dyn_scope)
                eq_expr = core_period.Eq(end_evt_scope, begin_scope)
                end_exprs.append(eq_expr)

            begin_expr = core_period.create_conjunction_from_exprs(begin_exprs)
            end_expr = core_period.create_conjunction_from_exprs(end_exprs)
            registry.add_period_def(None, None, begin_expr, end_expr, {}, {})

        # check that --period-captures name existing periods
        for name in name_to_begin_captures_exprs:
            if not registry.has_period_def(name):
                fmt = 'Cannot find period named "{}" for --period-captures ' \
                      'argument'
                self._cmdline_error(fmt.format(name))

    def _validate_transform_common_args(self):
        args = self._args
        refresh_period_ns = None
        if args.refresh is not None:
            try:
                refresh_period_ns = parse_utils.parse_duration(args.refresh)
            except ValueError as e:
                self._cmdline_error(str(e))

        self._analysis_conf = analysis.AnalysisConfig()
        self._analysis_conf.refresh_period = refresh_period_ns
        self._validate_transform_period_args(self._analysis_conf)

        if args.refresh is not None and not \
                self._analysis_conf.period_def_registry.is_empty:
            self._cmdline_error('Cannot specify --period* and --refresh '
                                'arguments at the same time')

        if args.cpu:
            self._analysis_conf.cpu_list = args.cpu.split(',')
            self._analysis_conf.cpu_list = [int(cpu) for cpu in
                                            self._analysis_conf.cpu_list]

        if args.debug:
            self._debug_mode = True

        # convert min/max args from µs to ns, if needed
        if hasattr(args, 'min') and args.min is not None:
            args.min *= 1000
            self._analysis_conf.min_duration = args.min
        if hasattr(args, 'max') and args.max is not None:
            args.max *= 1000
            self._analysis_conf.max_duration = args.max

        if hasattr(args, 'procname'):
            if args.procname:
                self._analysis_conf.proc_list = args.procname.split(',')

        if hasattr(args, 'tid'):
            if args.tid:
                self._analysis_conf.tid_list = args.tid.split(',')
                self._analysis_conf.tid_list = [int(tid) for tid in
                                                self._analysis_conf.tid_list]

        if hasattr(args, 'freq'):
            if args.freq_series:
                # implies uniform buckets
                args.freq_uniform = True

        if hasattr(args, 'freq') and args.freq_uniform:
            self._analysis_conf.uniform_min = {}
            self._analysis_conf.uniform_max = {}
            self._analysis_conf.uniform_step = {}

        if self._mi_mode:
            # print MI version if required
            if args.mi_version:
                print(mi.get_version_string())
                sys.exit(0)

            # print MI metadata if required
            if args.metadata:
                self._mi_print_metadata()
                sys.exit(0)

        # validate path argument (required at this point)
        if not args.path:
            self._cmdline_error('Please specify a trace path')

        if type(args.path) is list:
            args.path = args.path[0]

    def _validate_transform_args(self):
        pass

    def _parse_args(self):
        ap = argparse.ArgumentParser(description=self._DESC)

        # common arguments
        ap.add_argument('-r', '--refresh', type=str,
                        help='Refresh period, with optional units suffix '
                        '(default units: s)')
        ap.add_argument('--gmt', action='store_true',
                        help='Manipulate timestamps based on GMT instead '
                             'of local time')
        ap.add_argument('--skip-validation', action='store_true',
                        help='Skip the trace validation')
        ap.add_argument('--begin', type=str, help='start time: '
                                                  'hh:mm:ss[.nnnnnnnnn]')
        ap.add_argument('--end', type=str, help='end time: '
                                                'hh:mm:ss[.nnnnnnnnn]')
        ap.add_argument('--period', action='append', help='Period definition')
        ap.add_argument('--period-captures', action='append',
                        help='Period captures definition')
        ap.add_argument('--period-begin', type=str,
                        help='Analysis period start marker event name')
        ap.add_argument('--period-end', type=str,
                        help='Analysis period end marker event name '
                        '(requires --period-begin)')
        ap.add_argument('--period-begin-key', type=str,
                        help='Optional, list of event field names used to '
                        'match period markers (default: cpu_id)')
        ap.add_argument('--period-end-key', type=str,
                        help='Optional, list of event field names used to '
                        'match period marker. If none specified, use the same '
                        ' --period-begin-key')
        ap.add_argument('--period-key-value', type=str,
                        help='Optional, define a fixed key value to which a'
                        ' period must correspond to be considered.')
        ap.add_argument('--cpu', type=str,
                        help='Filter the results only for this list of '
                        'CPU IDs')
        ap.add_argument('--timerange', type=str, help='time range: '
                                                      '[begin,end]')
        ap.add_argument('--progress-use-size', action='store_true',
                        help='use trace size to approximate progress')
        ap.add_argument('--no-intersection', action='store_false',
                        dest='intersect_mode',
                        help='disable stream intersection mode')
        ap.add_argument('-V', '--version', action='version',
                        version='LTTng Analyses v{}'.format(self._VERSION))
        ap.add_argument('--debug', action='store_true',
                        help='Enable debug mode (or set {} environment '
                             'variable)'.format(self._DEBUG_ENV_VAR))
        ap.add_argument('--no-color', action='store_false', dest='color',
                        help='Disable colored output')

        # MI mode-dependent arguments
        if self._mi_mode:
            ap.add_argument('--mi-version', action='store_true',
                            help='Print MI version')
            ap.add_argument('--metadata', action='store_true',
                            help='Print analysis\' metadata')
            ap.add_argument('--test-compatibility', action='store_true',
                            help='Check if the provided trace is supported '
                                 'and exit')
            ap.add_argument('path', metavar='<path/to/trace>',
                            help='trace path', nargs='*')
            ap.add_argument('--output-progress', action='store_true',
                            help='Print progress indication lines')
        else:
            ap.add_argument('--no-progress', action='store_true',
                            help='Don\'t display the progress bar')
            ap.add_argument('path', metavar='<path/to/trace>',
                            help='trace path')

        # Used to add command-specific args
        self._add_arguments(ap)

        self._args = ap.parse_args()

        if self._mi_mode:
            # Compatiblity checking does not need to read the whole
            # trace, the caller should make sure there are no lost
            # events. At worst, they will be detected when the analysis
            # is actually run.
            if self._args.test_compatibility:
                self._args.skip_validation = True
            self._args.no_progress = True

            if self._args.output_progress:
                self._args.no_progress = False

        self._validate_transform_common_args()
        self._validate_transform_args()

    @staticmethod
    def _add_proc_filter_args(ap):
        ap.add_argument('--procname', type=str,
                        help='Filter the results only for this list of '
                        'process names')
        ap.add_argument('--tid', type=str,
                        help='Filter the results only for this list of TIDs')

    @staticmethod
    def _add_min_max_args(ap):
        ap.add_argument('--min', type=float,
                        help='Filter out durations shorter than min usec')
        ap.add_argument('--max', type=float,
                        help='Filter out durations longer than max usec')

    @staticmethod
    def _add_freq_args(ap, help=None):
        if not help:
            help = 'Output the frequency distribution'

        ap.add_argument('--freq', action='store_true', help=help)
        ap.add_argument('--freq-resolution', type=int, default=20,
                        help='Frequency distribution resolution '
                        '(default 20)')
        ap.add_argument('--freq-uniform', action='store_true',
                        help='Use a uniform resolution across distributions')
        ap.add_argument('--freq-series', action='store_true',
                        help='Consolidate frequency distribution histogram '
                        'as a single one')

    @staticmethod
    def _add_log_args(ap, help=None):
        if not help:
            help = 'Output the events in chronological order'

        ap.add_argument('--log', action='store_true', help=help)

    @staticmethod
    def _add_top_args(ap, help=None):
        if not help:
            help = 'Output the top results'

        ap.add_argument('--limit', type=int, default=10,
                        help='Limit to top X (default = 10)')
        ap.add_argument('--top', action='store_true', help=help)

    @staticmethod
    def _add_stats_args(ap, help=None):
        if not help:
            help = 'Output statistics'

        ap.add_argument('--stats', action='store_true', help=help)

    def _add_arguments(self, ap):
        pass

    def _process_date_args(self):
        def parse_date(date):
            try:
                ts = parse_utils.parse_trace_collection_date(
                    self._traces, date, self._args.gmt, self._handles
                )
            except ValueError as e:
                self._cmdline_error(str(e))

            return ts

        self._args.multi_day = trace_utils.is_multi_day_trace_collection(
            self._traces, self._handles)
        begin_ts = None
        end_ts = None

        if self._args.timerange:
            try:
                begin_ts, end_ts = (
                    parse_utils.parse_trace_collection_time_range(
                        self._traces, self._args.timerange,
                        self._args.gmt, self._handles)
                )
            except ValueError as e:
                self._cmdline_error(str(e))
        else:
            if self._args.begin:
                begin_ts = parse_date(self._args.begin)
            if self._args.end:
                end_ts = parse_date(self._args.end)

                # We have to check if timestamp_begin is None, which
                # it always is in older versions of babeltrace. In
                # that case, the test is simply skipped and an invalid
                # --end value will cause an empty analysis
                if self._ts_begin is not None and \
                   end_ts < self._ts_begin:
                    self._cmdline_error(
                        '--end timestamp before beginning of trace')

        self._analysis_conf.begin_ts = begin_ts
        self._analysis_conf.end_ts = end_ts

    def _create_analysis(self):
        notification_cbs = {
            analysis.AnalysisCallbackType.TICK_CB: self._analysis_tick_cb
        }

        self._analysis = self._ANALYSIS_CLASS(self.state, self._analysis_conf)
        self._analysis.register_notification_cbs(notification_cbs)

    def _create_automaton(self):
        self._automaton = automaton.Automaton()
        self.state = self._automaton.state

    def _analysis_tick_cb(self, period, end_ns):
        # No event was processed, just exit
        if end_ns is None:
            return
        self._analysis_tick(period, end_ns)

        if period is not None:
            # increment the number of effective ticks associated to
            # an existing period
            self._period_ticks += 1

    def _analysis_tick(self, period, end_ns):
        raise NotImplementedError()



# The MIT License (MIT)
#
# Copyright (C) 2016 - Julien Desfossez <jdesfossez@efficios.com>
#                      Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .analysis_test import AnalysisTest


class IrqTest(AnalysisTest):
    def write_trace(self):
        self.trace_writer.write_softirq_raise(1000, 1, 1)
        self.trace_writer.write_softirq_raise(1001, 3, 1)
        self.trace_writer.write_softirq_raise(1002, 1, 9)
        self.trace_writer.write_softirq_exit(1003, 0, 4)
        self.trace_writer.write_softirq_raise(1004, 3, 9)
        self.trace_writer.write_softirq_raise(1005, 3, 7)
        self.trace_writer.write_softirq_entry(1006, 3, 1)
        self.trace_writer.write_softirq_entry(1007, 1, 1)
        self.trace_writer.write_softirq_exit(1008, 1, 1)
        self.trace_writer.write_softirq_exit(1009, 3, 1)
        self.trace_writer.write_softirq_entry(1010, 1, 9)
        self.trace_writer.write_softirq_entry(1011, 3, 7)
        self.trace_writer.write_softirq_exit(1012, 1, 9)
        self.trace_writer.write_softirq_exit(1013, 3, 7)
        self.trace_writer.write_softirq_entry(1014, 3, 9)
        self.trace_writer.write_softirq_exit(1015, 3, 9)
        self.trace_writer.write_irq_handler_entry(1016, 0, 41, 'ahci')
        self.trace_writer.write_softirq_raise(1017, 0, 4)
        self.trace_writer.write_irq_handler_exit(1018, 0, 41, 1)
        self.trace_writer.write_softirq_entry(1019, 0, 4)
        self.trace_writer.write_softirq_exit(1020, 0, 4)
        self.trace_writer.write_irq_handler_entry(1021, 0, 41, 'ahci')
        self.trace_writer.write_softirq_raise(1022, 0, 4)
        self.trace_writer.write_irq_handler_exit(1023, 0, 41, 1)
        self.trace_writer.write_softirq_entry(1024, 0, 4)
        self.trace_writer.write_softirq_exit(1025, 0, 4)
        self.trace_writer.write_irq_handler_entry(1026, 0, 41, 'ahci')
        self.trace_writer.write_softirq_raise(1027, 0, 4)
        self.trace_writer.write_irq_handler_exit(1028, 0, 41, 1)
        self.trace_writer.write_softirq_entry(1029, 0, 4)
        self.trace_writer.write_softirq_exit(1030, 0, 4)
        self.trace_writer.write_irq_handler_entry(1031, 0, 41, 'ahci')
        self.trace_writer.write_softirq_raise(1032, 0, 4)
        self.trace_writer.write_irq_handler_exit(1033, 0, 41, 1)
        self.trace_writer.write_softirq_entry(1034, 0, 4)
        self.trace_writer.write_softirq_exit(1035, 0, 4)
        self.trace_writer.write_irq_handler_entry(1036, 0, 41, 'ahci')
        self.trace_writer.write_softirq_raise(1037, 0, 4)
        self.trace_writer.write_irq_handler_exit(1038, 0, 41, 1)
        self.trace_writer.write_softirq_entry(1039, 0, 4)
        self.trace_writer.write_softirq_exit(1040, 0, 4)
        self.trace_writer.write_irq_handler_entry(1041, 0, 41, 'ahci')
        self.trace_writer.write_softirq_raise(1042, 0, 4)
        self.trace_writer.write_irq_handler_exit(1043, 0, 41, 1)
        self.trace_writer.write_softirq_entry(1044, 0, 4)
        self.trace_writer.write_softirq_exit(1045, 0, 4)
        self.trace_writer.flush()

    def test_irqstats(self):
        test_name = 'irqstats'
        expected = self.get_expected_output(test_name)
        result = self.get_cmd_output('lttng-irqstats',
                                     options='--no-intersection')

        self._assertMultiLineEqual(result, expected, test_name)

    def test_irqlog(self):
        test_name = 'irqlog'
        expected = self.get_expected_output(test_name)
        result = self.get_cmd_output('lttng-irqlog',
                                     options='--no-intersection')

        self._assertMultiLineEqual(result, expected, test_name)
